<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 08/25/2003 *
 * ===============                              Last change: 11/29/2005 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : functions.php                                    *
 * -------------------------------------------------------------------- *
 * Short description : Many non-MySQL functions (also file access)      *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Viele Nicht-MySQL-Funktionen (auch Dateizugriff) *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * Needs to be in all Files and every File needs "svn propset           *
 * svn:keywords Date Revision" (autoprobset!) at least!!!!!!            *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// Output HTML code directly or 'render' it. You addionally switch the new-line character off
function outputHtml ($htmlCode, $newLine = true) {
	// Transfer username
	$username = getMessage('USERNAME_UNKNOWN');
	if (isset($GLOBALS['username'])) $username = getUsername();

	// Do we have HTML-Code here?
	if (!empty($htmlCode)) {
		// Yes, so we handle it as you have configured
		switch (getConfig('OUTPUT_MODE')) {
			case 'render':
				// That's why you don't need any \n at the end of your HTML code... :-)
				if (getPhpCaching() == 'on') {
					// Output into PHP's internal buffer
					outputRawCode($htmlCode);

					// That's why you don't need any \n at the end of your HTML code... :-)
					if ($newLine === true) print("\n");
				} else {
					// Render mode for old or lame servers...
					$GLOBALS['output'] .= $htmlCode;

					// That's why you don't need any \n at the end of your HTML code... :-)
					if ($newLine === true) $GLOBALS['output'] .= "\n";
				}
				break;

			case 'direct':
				// If we are switching from render to direct output rendered code
				if ((!empty($GLOBALS['output'])) && (getPhpCaching() != 'on')) { outputRawCode($GLOBALS['output']); $GLOBALS['output'] = ''; }

				// The same as above... ^
				outputRawCode($htmlCode);
				if ($newLine) print("\n");
				break;

			default:
				// Huh, something goes wrong or maybe you have edited config.php ???
				app_die(__FUNCTION__, __LINE__, '<strong>{--FATAL_ERROR--}:</strong> {--LANG_NO_RENDER_DIRECT--}');
				break;
		} // END - switch
	} elseif ((getPhpCaching() == 'on') && (isset($GLOBALS['footer_sent'])) && ($GLOBALS['footer_sent'] == 1)) {
		// Headers already sent?
		if (headers_sent()) {
			// Log this error
			logDebugMessage(__FUNCTION__, __LINE__, 'Headers already sent! We need debug backtrace here.');

			// Trigger an user error
			debug_report_bug('Headers are already sent!');
		} // END - if

		// Output cached HTML code
		$GLOBALS['output'] = ob_get_contents();

		// Clear output buffer for later output if output is found
		if (!empty($GLOBALS['output'])) {
			clearOutputBuffer();
		} // END - if

		// Send HTTP header
		sendHeader('HTTP/1.1 200');

		// Used later
		$now = gmdate('D, d M Y H:i:s') . ' GMT';

		// General headers for no caching
		sendHeader('Expired: ' . $now); // RFC2616 - Section 14.21
		sendHeader('Last-Modified: ' . $now);
		sendHeader('Cache-Control: no-store, no-cache, must-revalidate, pre-check=0, post-check=0, max-age=0'); // HTTP/1.1
		sendHeader('Pragma: no-cache'); // HTTP/1.0
		sendHeader('Connection: Close');
		sendHeader('Content-Type: ' . getContentType() . '; charset=UTF-8');
		sendHeader('Content-language: ' . getLanguage());

		// Extension 'rewrite' installed?
		if ((isExtensionActive('rewrite')) && (getOutputMode() != 1)) {
			$GLOBALS['output'] = rewriteLinksInCode($GLOBALS['output']);
		} // END - if

		// Init counter
		$cnt = 0;

		// Compile and run finished rendered HTML code
		while (((strpos($GLOBALS['output'], '{--') > 0) || (strpos($GLOBALS['output'], '{!') > 0) || (strpos($GLOBALS['output'], '{?') > 0)) && ($cnt < 3)) {
			// Prepare the content and eval() it...
			$content = array();
			$newContent = '';

			// Compile it
			$eval = "\$newContent = \"".compileCode(smartAddSlashes($GLOBALS['output']))."\";";
			eval($eval);

			// Was that eval okay?
			if (empty($newContent)) {
				// Something went wrong!
				debug_report_bug('Evaluation error:<pre>' . linenumberCode($eval) . '</pre>');
			} // END - if
			$GLOBALS['output'] = $newContent;

			// Count round
			$cnt++;
		} // END - while

		// Output code here, DO NOT REMOVE! ;-)
		outputRawCode($GLOBALS['output']);
	} elseif ((getConfig('OUTPUT_MODE') == 'render') && (!empty($GLOBALS['output']))) {
		// Rewrite links when rewrite extension is active
		if ((isExtensionActive('rewrite')) && (getOutputMode() != 1)) {
			$GLOBALS['output'] = rewriteLinksInCode($GLOBALS['output']);
		} // END - if

		// Compile and run finished rendered HTML code
		while (strpos($GLOBALS['output'], '{!') > 0) {
			eval("\$GLOBALS['output'] = \"".compileCode(smartAddSlashes($GLOBALS['output']))."\";");
		} // END - while

		// Output code here, DO NOT REMOVE! ;-)
		outputRawCode($GLOBALS['output']);
	}
}

// Output the raw HTML code
function outputRawCode ($htmlCode) {
	// Output stripped HTML code to avoid broken JavaScript code, etc.
	print(stripslashes(stripslashes($htmlCode)));

	// Flush the output if only getPhpCaching() is not 'on'
	if (getPhpCaching() != 'on') {
		// Flush it
		flush();
	} // END - if
}

// Init fatal message array
function initFatalMessages () {
	$GLOBALS['fatal_messages'] = array();
}

// Getter for whole fatal error messages
function getFatalArray () {
	return $GLOBALS['fatal_messages'];
}

// Add a fatal error message to the queue array
function addFatalMessage ($F, $L, $message, $extra='') {
	if (is_array($extra)) {
		// Multiple extras for a message with masks
		$message = call_user_func_array('sprintf', $extra);
	} elseif (!empty($extra)) {
		// $message is text with a mask plus extras to insert into the text
		$message = sprintf($message, $extra);
	}

	// Add message to $GLOBALS['fatal_messages']
	$GLOBALS['fatal_messages'][] = $message;

	// Log fatal messages away
	debug_report_bug($message);
	logDebugMessage($F, $L, " message={$message}");
}

// Getter for total fatal message count
function getTotalFatalErrors () {
	// Init coun
	$count = 0;

	// Do we have at least the first entry?
	if (!empty($GLOBALS['fatal_messages'][0])) {
		// Get total count
		$count = count($GLOBALS['fatal_messages']);
	} // END - if

	// Return value
	return $count;
}

// Load a template file and return it's content (only it's name; do not use ' or ")
function loadTemplate ($template, $return=false, $content=array()) {
	// @TODO Remove this sanity-check if all is fine
	if (!is_bool($return)) debug_report_bug('return is not bool (' . gettype($return) . ')');

	// @TODO Try to rewrite all $DATA to $content
	global $DATA;

	// Do we have cache?
	if (!isset($GLOBALS['template_eval'][$template])) {
		// Add more variables which you want to use in your template files
		$username = getUsername();

		// Make all template names lowercase
		$template = strtolower($template);

		// Count the template load
		incrementConfigEntry('num_templates');

		// Init some data
		$ret = '';
		if (empty($GLOBALS['refid'])) $GLOBALS['refid'] = 0;

		// Generate date/time string
		$date_time = generateDateTime(time(), 1);

		// Is content an array
		if (is_array($content)) $content['date_time'] = $date_time;

		// @DEPRECATED Try to rewrite the if() condition
		if ($template == 'member_support_form') {
			// Support request of a member
			$result = SQL_QUERY_ESC("SELECT `userid`, `gender`, `surname`, `family`, `email` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
				array(getUserId()), __FUNCTION__, __LINE__);

			// Is content an array?
			if (is_array($content)) {
				// Merge data
				$content = merge_array($content, SQL_FETCHARRAY($result));

				// Translate gender
				$content['gender'] = translateGender($content['gender']);
			} else {
				// @DEPRECATED
				// @TODO Find all templates which are using these direct variables and rewrite them.
				// @TODO After this step is done, this else-block is history
				list($gender, $surname, $family, $email) = SQL_FETCHROW($result);

				// Translate gender
				$gender = translateGender($gender);
				logDebugMessage(__FUNCTION__, __LINE__, sprintf("DEPRECATION-WARNING: content is not array [%s], template=%s.", gettype($content), $template));
			}

			// Free result
			SQL_FREERESULT($result);
		} // END - if

		// Base directory
		$basePath = sprintf("%stemplates/%s/html/", getConfig('PATH'), getLanguage());
		$mode = '';

		// Check for admin/guest/member templates
		if (substr($template, 0, 6) == 'admin_') {
			// Admin template found
			$mode = 'admin/';
		} elseif (substr($template, 0, 6) == 'guest_') {
			// Guest template found
			$mode = 'guest/';
		} elseif (substr($template, 0, 7) == 'member_') {
			// Member template found
			$mode = 'member/';
		} elseif (substr($template, 0, 8) == 'install_') {
			// Installation template found
			$mode = 'install/';
		} elseif (substr($template, 0, 4) == 'ext_') {
			// Extension template found
			$mode = 'ext/';
		} elseif (substr($template, 0, 3) == 'la_') {
			// 'Logical-area' template found
			$mode = 'la/';
		} elseif (substr($template, 0, 3) == 'js_') {
			// JavaScript template found
			$mode = 'js/';
		} elseif (substr($template, 0, 5) == 'menu_') {
			// Menu template found
			$mode = 'menu/';
		} else {
			// Test for extension
			$test = substr($template, 0, strpos($template, '_'));

			// Probe for valid extension name
			if (isExtensionNameValid($test)) {
				// Set extra path to extension's name
				$mode = $test . '/';
			} // END - if
		}

		////////////////////////
		// Generate file name //
		////////////////////////
		$FQFN = $basePath . $mode . $template . '.tpl';

		if ((isWhatSet()) && ((strpos($template, '_header') > 0) || (strpos($template, '_footer') > 0)) && (($mode == 'guest/') || ($mode == 'member/') || ($mode == 'admin/'))) {
			// Select what depended header/footer template file for admin/guest/member area
			$file2 = sprintf("%s%s%s_%s.tpl",
				$basePath,
				$mode,
				$template,
				getWhat()
			);

			// Probe for it...
			if (isFileReadable($file2)) $FQFN = $file2;

			// Remove variable from memory
			unset($file2);
		} // END - if

		// Does the special template exists?
		if (!isFileReadable($FQFN)) {
			// Reset to default template
			$FQFN = $basePath . $template . '.tpl';
		} // END - if

		// Now does the final template exists?
		if (isFileReadable($FQFN)) {
			// The local file does exists so we load it. :)
			$GLOBALS['tpl_content'] = readFromFile($FQFN);

			// Replace ' to our own chars to preventing them being quoted
			while (strpos($GLOBALS['tpl_content'], "'") !== false) { $GLOBALS['tpl_content'] = str_replace("'", '{QUOT}', $GLOBALS['tpl_content']); }

			// Do we have to compile the code?
			$ret = '';
			if ((strpos($GLOBALS['tpl_content'], '$') !== false) || (strpos($GLOBALS['tpl_content'], '{--') !== false) || (strpos($GLOBALS['tpl_content'], '{!') !== false) || (strpos($GLOBALS['tpl_content'], '{?') !== false)) {
				// Normal HTML output?
				if ($GLOBALS['output_mode'] == 0) {
					// Add surrounding HTML comments to help finding bugs faster
					$ret = "<!-- Template " . $template . " - Start -->\n" . $GLOBALS['tpl_content'] . "<!-- Template " . $template . " - End -->\n";

					// Prepare eval() command
					$eval = '$ret = "' . compileCode(smartAddSlashes($GLOBALS['tpl_content'])) . '";';
				} else {
					// Prepare eval() command
					$eval = '$ret = "' . compileCode(smartAddSlashes($GLOBALS['tpl_content'])) . '";';
				}
			} else {
				// Add surrounding HTML comments to help finding bugs faster
				$ret = "<!-- Template " . $template . " - Start -->\n" . $GLOBALS['tpl_content'] . "<!-- Template " . $template . " - End -->\n";
				$eval = '$ret = "' . smartAddSlashes($ret) . '";';
			} // END - if

			// Cache the eval() command here
			$GLOBALS['template_eval'][$template] = $eval;

			// Eval the code
			eval($GLOBALS['template_eval'][$template]);
		} else {
			// No file!
			$GLOBALS['template_eval'][$template] = '404';
		}
	} elseif (((isAdmin()) || ((isInstalling()) && (!isInstalled()))) && ($GLOBALS['template_eval'][$template] == '404')) {
		// Only admins shall see this warning or when installation mode is active
		$ret = '<br /><span class=\\"guest_failed\\">{--TEMPLATE_404--}</span><br />
(' . $template . ')<br />
<br />
{--TEMPLATE_CONTENT--}
<pre>' . print_r($content, true) . '</pre>
{--TEMPLATE_DATA--}
<pre>' . print_r($DATA, true) . '</pre>
<br /><br />\";';
	} else {
		// Eval the code
		eval($GLOBALS['template_eval'][$template]);
	}

	// Do we have some content to output or return?
	if (!empty($ret)) {
		// Not empty so let's put it out! ;)
		if ($return === true) {
			// Return the HTML code
			return $ret;
		} else {
			// Output direct
			outputHtml($ret);
		}
	} elseif (isDebugModeEnabled()) {
		// Warning, empty output!
		return 'E:' . $template . ',content=<pre>' . print_r($content, true) . '</pre>';
	}
}

// Loads an email template and compiles it
function loadEmailTemplate ($template, $content = array(), $UID = 0) {
	global $DATA;

	// Our configuration is kept non-global here
	$_CONFIG = getConfigArray();

	// Make sure all template names are lowercase!
	$template = strtolower($template);

	// Default 'nickname' if extension is not installed
	$nick = '---';

	// Prepare IP number and User Agent
	$REMOTE_ADDR     = detectRemoteAddr();
	$HTTP_USER_AGENT = detectUserAgent();

	// Default admin
	$ADMIN = getConfig('MAIN_TITLE');

	// Is the admin logged in?
	if (isAdmin()) {
		// Get admin id
		$adminId = getCurrentAdminId();

		// Load Admin data
		$ADMIN = getAdminEmail($adminId);
	} // END - if

	// Neutral email address is default
	$email = getConfig('WEBMASTER');

	// Expiration in a nice output format
	// NOTE: Use $content[expiration] in your templates instead of $EXPIRATION
	if (getConfig('auto_purge') == 0) {
		// Will never expire!
		$EXPIRATION = getMessage('MAIL_WILL_NEVER_EXPIRE');
	} else {
		// Create nice date string
		$EXPIRATION = createFancyTime(getConfig('auto_purge'));
	}

	// Is content an array?
	if (is_array($content)) {
		// Add expiration to array, $EXPIRATION is now deprecated!
		$content['expiration'] = $EXPIRATION;
	} // END - if

	// Load user's data
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):UID={$UID},template={$template},content[]=".gettype($content)."<br />");
	if (($UID > 0) && (is_array($content))) {
		// If nickname extension is installed, fetch nickname as well
		if (isExtensionActive('nickname')) {
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):NICKNAME!<br />");
			// Load nickname
			$result = SQL_QUERY_ESC("SELECT `surname`, `family`, `gender`, `email`, `nickname` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
				array(bigintval($UID)), __FUNCTION__, __LINE__);
		} else {
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):NO-NICK!<br />");
			/// Load normal data
			$result = SQL_QUERY_ESC("SELECT `surname`, `family`, `gender`, `email` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
				array(bigintval($UID)), __FUNCTION__, __LINE__);
		}

		// Fetch and merge data
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):content()=".count($content)." - PRE<br />");
		$content = merge_array($content, SQL_FETCHARRAY($result));
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):content()=".count($content)." - AFTER<br />");

		// Free result
		SQL_FREERESULT($result);
	} // END - if

	// Translate M to male or F to female if present
	if (isset($content['gender'])) $content['gender'] = translateGender($content['gender']);

	// Overwrite email from data if present
	if (isset($content['email'])) $email = $content['email'];

	// Store email for some functions in global data array
	$DATA['email'] = $email;

	// Base directory
	$basePath = sprintf("%stemplates/%s/emails/", getConfig('PATH'), getLanguage());

	// Check for admin/guest/member templates
	if (substr($template, 0, 6) == 'admin_') {
		// Admin template found
		$FQFN = $basePath.'admin/' . $template.'.tpl';
	} elseif (substr($template, 0, 6) == 'guest_') {
		// Guest template found
		$FQFN = $basePath.'guest/' . $template.'.tpl';
	} elseif (substr($template, 0, 7) == 'member_') {
		// Member template found
		$FQFN = $basePath.'member/' . $template.'.tpl';
	} else {
		// Test for extension
		$test = substr($template, 0, strpos($template, '_'));
		if (isExtensionNameValid($test)) {
			// Set extra path to extension's name
			$FQFN = $basePath . $test.'/' . $template.'.tpl';
		} else {
			// No special filename
			$FQFN = $basePath . $template.'.tpl';
		}
	}

	// Does the special template exists?
	if (!isFileReadable($FQFN)) {
		// Reset to default template
		$FQFN = $basePath . $template.'.tpl';
	} // END - if

	// Now does the final template exists?
	$newContent = '';
	if (isFileReadable($FQFN)) {
		// The local file does exists so we load it. :)
		$GLOBALS['tpl_content'] = readFromFile($FQFN);
		$GLOBALS['tpl_content'] = SQL_ESCAPE($GLOBALS['tpl_content']);

		// Run code
		$GLOBALS['tpl_content'] = "\$newContent = decodeEntities(\"".compileCode($GLOBALS['tpl_content'])."\");";
		eval($GLOBALS['tpl_content']);
	} elseif (!empty($template)) {
		// Template file not found!
		$newContent = "{--TEMPLATE_404--}: " . $template."<br />
{--TEMPLATE_CONTENT--}
<pre>".print_r($content, true)."</pre>
{--TEMPLATE_DATA--}
<pre>".print_r($DATA, true)."</pre>
<br /><br />";

		// Debug mode not active? Then remove the HTML tags
		if (!isDebugModeEnabled()) $newContent = secureString($newContent);
	} else {
		// No template name supplied!
		$newContent = getMessage('NO_TEMPLATE_SUPPLIED');
	}

	// Is there some content?
	if (empty($newContent)) {
		// Compiling failed
		$newContent = "Compiler error for template {$template}!\nUncompiled content:\n" . $GLOBALS['tpl_content'];
		// Add last error if the required function exists
		if (function_exists('error_get_last')) $newContent .= "\n--------------------------------------\nDebug:\n".print_r(error_get_last(), true)."--------------------------------------\nPlease don't alter these informations!\nThanx.";
	} // END - if

	// Remove content and data
	unset($content);
	unset($DATA);

	// Compile the code and eval it
	$eval = '$newContent = "' . compileCode(smartAddSlashes($newContent)) . '";';
	eval($eval);

	// Return content
	return $newContent;
}

// Send mail out to an email address
function sendEmail ($toEmail, $subject, $message, $isHtml = 'N', $mailHeader = '') {
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):TO={$toEmail},SUBJECT={$subject}<br />");

	// Compile subject line (for POINTS constant etc.)
	eval("\$subject = decodeEntities(\"".compileCode(smartAddSlashes($subject))."\");");

	// Set from header
	if ((!eregi('@', $toEmail)) && ($toEmail > 0)) {
		// Value detected, is the message extension installed?
		// @TODO Extension 'msg' does not exist
		if (isExtensionActive('msg')) {
			ADD_MESSAGE_TO_BOX($toEmail, $subject, $message, $isHtml);
			return;
		} else {
			// Load email address
			$result_email = SQL_QUERY_ESC("SELECT `email` FROM `{?_MYSQL_PREFIX?}_user_data` WHERE `userid`=%s LIMIT 1",
				array(bigintval($toEmail)), __FUNCTION__, __LINE__);
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):numRows=".SQL_NUMROWS($result_email)."<br />");

			// Does the user exist?
			if (SQL_NUMROWS($result_email)) {
				// Load email address
				list($toEmail) = SQL_FETCHROW($result_email);
			} else {
				// Set webmaster
				$toEmail = getConfig('WEBMASTER');
			}

			// Free result
			SQL_FREERESULT($result_email);
		}
	} elseif ($toEmail == 0) {
		// Is the webmaster!
		$toEmail = getConfig('WEBMASTER');
	}
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):TO={$toEmail}<br />");

	// Check for PHPMailer or debug-mode
	if (!checkPhpMailerUsage()) {
		// Not in PHPMailer-Mode
		if (empty($mailHeader)) {
			// Load email header template
			$mailHeader = loadEmailTemplate('header');
		} else {
			// Append header
			$mailHeader .= loadEmailTemplate('header');
		}
	} elseif (isDebugModeEnabled()) {
		if (empty($mailHeader)) {
			// Load email header template
			$mailHeader = loadEmailTemplate('header');
		} else {
			// Append header
			$mailHeader .= loadEmailTemplate('header');
		}
	}

	// Compile "TO"
	eval("\$toEmail = \"".compileCode(smartAddSlashes($toEmail))."\";");

	// Compile "MSG"
	eval("\$message = \"".compileCode(smartAddSlashes($message))."\";");

	// Fix HTML parameter (default is no!)
	if (empty($isHtml)) $isHtml = 'N';
	if (isDebugModeEnabled()) {
		// In debug mode we want to display the mail instead of sending it away so we can debug this part
		outputHtml('<pre>
Headers : ' . htmlentities(trim($mailHeader)) . '
To      : ' . $toEmail . '
Subject : ' . $subject . '
Message : ' . $message . '
</pre>');
	} elseif (($isHtml == 'Y') && (isExtensionActive('html_mail'))) {
		// Send mail as HTML away
		sendHtmlEmail($toEmail, $subject, $message, $mailHeader);
	} elseif (!empty($toEmail)) {
		// Send Mail away
		sendRawEmail($toEmail, $subject, $message, $mailHeader);
	} elseif ($isHtml != 'Y') {
		// Problem found!
		sendRawEmail(getConfig('WEBMASTER'), '[PROBLEM:]' . $subject, $message, $mailHeader);
	}
}

// Check if legacy or PHPMailer command
// @TODO Rewrite this to an extension 'smtp'
// @private
function checkPhpMailerUsage() {
	return ((getConfig('SMTP_HOSTNAME') != '') && (getConfig('SMTP_USER') != ''));
}

// Send out a raw email with PHPMailer class or legacy mail() command
function sendRawEmail ($toEmail, $subject, $message, $from) {
	// Shall we use PHPMailer class or legacy mode?
	if (checkPhpMailerUsage()) {
		// Use PHPMailer class with SMTP enabled
		loadIncludeOnce('inc/phpmailer/class.phpmailer.php');
		loadIncludeOnce('inc/phpmailer/class.smtp.php');

		// get new instance
		$mail = new PHPMailer();
		$mail->PluginDir  = sprintf("%sinc/phpmailer/", getConfig('PATH'));

		$mail->IsSMTP();
		$mail->SMTPAuth   = true;
		$mail->Host       = getConfig('SMTP_HOSTNAME');
		$mail->Port       = 25;
		$mail->Username   = getConfig('SMTP_USER');
		$mail->Password   = getConfig('SMTP_PASSWORD');
		if (empty($from)) {
			$mail->From = getConfig('WEBMASTER');
		} else {
			$mail->From = $from;
		}
		$mail->FromName   = getConfig('MAIN_TITLE');
		$mail->Subject    = $subject;
		if ((isExtensionActive('html_mail')) && (secureString($message) != $message)) {
			$mail->Body       = $message;
			$mail->AltBody    = 'Your mail program required HTML support to read this mail!';
			$mail->WordWrap   = 70;
			$mail->IsHTML(true);
		} else {
			$mail->Body       = decodeEntities($message);
		}
		$mail->AddAddress($toEmail, '');
		$mail->AddReplyTo(getConfig('WEBMASTER'), getConfig('MAIN_TITLE'));
		$mail->AddCustomHeader('Errors-To:' . getConfig('WEBMASTER'));
		$mail->AddCustomHeader('X-Loop:' . getConfig('WEBMASTER'));
		$mail->Send();
	} else {
		// Use legacy mail() command
		mail($toEmail, $subject, decodeEntities($message), $from);
	}
}

// Generate a password in a specified length or use default password length
function generatePassword ($length = 0) {
	// Auto-fix invalid length of zero
	if ($length == 0) $length = getConfig('pass_len');

	// Initialize array with all allowed chars
	$ABC = explode(',', 'a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z,A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P,Q,R,S,T,U,V,W,X,Y,Z,0,1,2,3,4,5,6,7,8,9,-,+,_,/,.');

	// Start creating password
	$PASS = '';
	for ($i = 0; $i < $length; $i++) {
		$PASS .= $ABC[mt_rand(0, count($ABC) -1)];
	} // END - for

	// When the size is below 40 we can also add additional security by scrambling
	// it. Otherwise we may corrupt hashes
	if (strlen($PASS) <= 40) {
		// Also scramble the password
		$PASS = scrambleString($PASS);
	} // END - if

	// Return the password
	return $PASS;
}

// Generates a human-readable timestamp from the Uni* stamp
function generateDateTime ($time, $mode = 0) {
	// Filter out numbers
	$time = bigintval($time);

	// If the stamp is zero it mostly didn't "happen"
	if ($time == 0) {
		// Never happend
		return getMessage('NEVER_HAPPENED');
	} // END - if

	switch (getLanguage()) {
		case 'de': // German date / time format
			switch ($mode) {
				case 0: $ret = date("d.m.Y \u\m H:i \U\h\\r", $time); break;
				case 1: $ret = strtolower(date('d.m.Y - H:i', $time)); break;
				case 2: $ret = date('d.m.Y|H:i', $time); break;
				case 3: $ret = date('d.m.Y', $time); break;
				default:
					logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
					break;
			}
			break;

		default: // Default is the US date / time format!
			switch ($mode) {
				case 0: $ret = date('r', $time); break;
				case 1: $ret = date('Y-m-d - g:i A', $time); break;
				case 2: $ret = date('y-m-d|H:i', $time); break;
				case 3: $ret = date('y-m-d', $time); break;
				default:
					logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid date mode %s detected.", $mode));
					break;
			} // END - switch
	} // END - switch

	// Return result
	return $ret;
}

// Translates Y/N to yes/no
function translateYesNo ($yn) {
	// Default
	$translated = '??? (' . $yn . ')';
	switch ($yn) {
		case 'Y': $translated = getMessage('YES'); break;
		case 'N': $translated = getMessage('NO'); break;
		default:
			// Log unknown value
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown value %s. Expected Y/N!", $yn));
			break;
	}

	// Return it
	return $translated;
}

// Translates the "pool type" into human-readable
function translatePoolType ($type) {
	// Default?type is unknown
	$translated = sprintf(getMessage('POOL_TYPE_UNKNOWN'), $type);

	// Generate constant
	$constName = sprintf("POOL_TYPE_%s", $type);

	// Does it exist?
	if (isMessageIdValid($constName)) {
		// Then use it
		$translated = getMessage($constName);
	} // END - if

	// Return "translation"
	return $translated;
}

// Translates the american decimal dot into a german comma
function translateComma ($dotted, $cut = true, $max = 0) {
	// Default is 3 you can change this in admin area "Misc -> Misc Options"
	if (!isConfigEntrySet('max_comma')) setConfigEntry('max_comma', 3);

	// Use from config is default
	$maxComma = getConfig('max_comma');

	// Use from parameter?
	if ($max > 0) $maxComma = $max;

	// Cut zeros off?
	if (($cut === true) && ($max == 0)) {
		// Test for commata if in cut-mode
		$com = explode('.', $dotted);
		if (count($com) < 2) {
			// Don't display commatas even if there are none... ;-)
			$maxComma = 0;
		}
	} // END - if

	// Debug log
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, "dotted={$dotted},maxComma={$maxComma}");

	// Translate it now
	switch (getLanguage()) {
		case 'de': // German language
			$dotted = number_format($dotted, $maxComma, ',', '.');
			break;

		default: // All others
			$dotted = number_format($dotted, $maxComma, '.', ',');
			break;
	}

	// Return translated value
	return $dotted;
}

// Translate Uni*-like gender to human-readable
function translateGender ($gender) {
	// Default
	$ret = '!' . $gender . '!';

	// Male/female or company?
	switch ($gender) {
		case 'M': $ret = getMessage('GENDER_M'); break;
		case 'F': $ret = getMessage('GENDER_F'); break;
		case 'C': $ret = getMessage('GENDER_C'); break;
		default:
			// Log unknown gender
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown gender %s detected.", $gender));
			break;
	}

	// Return translated gender
	return $ret;
}

// "Translates" the user status
function translateUserStatus ($status) {
	// Generate message depending on status
	switch ($status) {
		case 'UNCONFIRMED':
		case 'CONFIRMED':
		case 'LOCKED':
			$ret = getMessage(sprintf("ACCOUNT_%s", $status));
			break;

		case '':
		case null:
			$ret = getMessage('ACCOUNT_DELETED');
			break;

		default:
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown status %s detected.", $status));
			$ret = sprintf(getMessage('UNKNOWN_STATUS'), $status);
			break;
	} // END - switch

	// Return it
	return $ret;
}

// Generates an URL for the dereferer
function generateDerefererUrl ($URL) {
	// Don't de-refer our own links!
	if (substr($URL, 0, strlen(getConfig('URL'))) != getConfig('URL')) {
		// De-refer this link
		$URL = '{?URL?}/modules.php?module=loader&amp;url=' . encodeString(compileUriCode($URL));
	} // END - if

	// Return link
	return $URL;
}

// Generates an URL for the frametester
function generateFrametesterUrl ($URL) {
	// Prepare frametester URL
	$frametesterUrl = sprintf("{?URL?}/modules.php?module=frametester&amp;url=%s",
		encodeString(compileUriCode($URL))
	);

	// Return the new URL
	return $frametesterUrl;
}

// Count entries from e.g. a selection box
function countSelection ($array) {
	// Integrity check
	if (!is_array($array)) {
		// Not an array!
		debug_report_bug(__FUNCTION__.': No array provided.');
	} // END - if

	// Init count
	$ret = 0;

	// Count all entries
	foreach ($array as $key => $selected) {
		// Is it checked?
		if (!empty($selected)) $ret++;
	} // END - foreach

	// Return counted selections
	return $ret;
}

// Generate XHTML code for the CAPTCHA
function generateCaptchaCode ($code, $type, $DATA, $userid) {
	return '<img border="0" alt="Code ' . $code . '" src="{?URL?}/mailid_top.php?userid=' . $userid . '&amp;' . $type . '=' . $DATA . '&amp;mode=img&amp;code=' . $code . '" />';
}

// Generates a timestamp (some wrapper for mktime())
function makeTime ($hours, $minutes, $seconds, $stamp) {
	// Extract day, month and year from given timestamp
	$days   = date('d', $stamp);
	$months = date('m', $stamp);
	$years  = date('Y', $stamp);

	// Create timestamp for wished time which depends on extracted date
	return mktime(
		$hours,
		$minutes,
		$seconds,
		$months,
		$days,
		$years
	);
}

// Redirects to an URL and if neccessarry extends it with own base URL
function redirectToUrl ($URL) {
	// Compile out URI codes
	$URL = compileUriCode($URL);

	// Check if http(s):// is there
	if ((substr($URL, 0, 7) != 'http://') && (substr($URL, 0, 8) != 'https://')) {
		// Make all URLs full-qualified
		$URL = getConfig('URL') . '/' . $URL;
	} // END - if

	// Three different debug ways...
	//* DEBUG: */ debug_report_bug(sprintf("%s[%s:] URL=%s", __FUNCTION__, __LINE__, $URL));
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $URL);
	//* DEBUG: */ die($URL);

	// Default 'rel' value is external, nofollow is evil from Google and hurts the Internet
	$rel = ' rel="external"';

	// Do we have internal or external URL?
	if (substr($URL, 0, strlen(getConfig('URL'))) == getConfig('URL')) {
		// Own (=internal) URL
		$rel = '';
	} // END - if

	// Get output buffer
	$GLOBALS['output'] = ob_get_contents();

	// Clear it only if there is content
	if (!empty($GLOBALS['output'])) {
		clearOutputBuffer();
	} // END - if

	// Simple probe for bots/spiders from search engines
	if ((strpos(detectUserAgent(), 'spider') !== false) || (strpos(detectUserAgent(), 'bot') !== false)) {
		// Secure the URL against bad things such als HTML insertions and so on...
		$URL = secureString($URL);

		// Output new location link as anchor
		outputHtml('<a href="' . $URL . '"' . $rel . '>' . $URL . '</a>');
	} elseif (!headers_sent()) {
		// Load URL when headers are not sent
		//* DEBUG: */ debug_report_bug("URL={$URL}");
		sendHeader('Location: '.str_replace('&amp;', '&', $URL));
	} else {
		// Output error message
		loadInclude('inc/header.php');
		loadTemplate('redirect_url', false, str_replace('&amp;', '&', $URL));
		loadInclude('inc/footer.php');
	}

	// Shut the mailer down here
	shutdown();
}

// Wrapper for redirectToUrl but URL comes from a configuration entry
function redirectToConfiguredUrl ($configEntry) {
	// Get the URL
	$URL = getConfig($configEntry);

	// Is this URL set?
	if (is_null($URL)) {
		// Then abort here
		debug_report_bug(sprintf("Configuration entry %s is not set!", $configEntry));
	} // END - if

	// Load the URL
	redirectToUrl($URL);
}

// Compiles the given HTML/mail code
function compileCode ($code, $simple = false, $constants = true, $full = true) {
	// Is the code a string?
	if (!is_string($code)) {
		// Silently return it
		return $code;
	} // END - if

	// Init replacement-array with full security characters
	$secChars = $GLOBALS['security_chars'];

	// Select smaller set of chars to replace when we e.g. want to compile URLs
	if ($full === false) $secChars = $GLOBALS['url_chars'];

	// Compile more through a filter
	$code = runFilterChain('compile_code', $code);

	// Compile constants
	if ($constants === true) {
		// BEFORE 0.2.1 : Language and data constants
		// WITH 0.2.1+  : Only language constants
		$code = str_replace('{--', "\".getMessage('", str_replace('--}', "').\"", $code));

		// BEFORE 0.2.1 : Not used
		// WITH 0.2.1+  : Data constants
		$code = str_replace('{!', "\".constant('", str_replace("!}", "').\"", $code));
	} // END - if

	// Compile QUOT and other non-HTML codes
	foreach ($secChars['to'] as $k => $to) {
		// Do the reversed thing as in inc/libs/security_functions.php
		$code = str_replace($to, $secChars['from'][$k], $code);
	} // END - foreach

	// But shall I keep simple quotes for later use?
	if ($simple) $code = str_replace("'", '{QUOT}', $code);

	// Find $content[bla][blub] entries
	preg_match_all('/\$(content|GLOBALS|DATA)((\[([a-zA-Z0-9-_]+)\])*)/', $code, $matches);

	// Are some matches found?
	if ((count($matches) > 0) && (count($matches[0]) > 0)) {
		// Replace all matches
		$matchesFound = array();
		foreach ($matches[0] as $key => $match) {
			// Fuzzy look has failed by default
			$fuzzyFound = false;

			// Fuzzy look on match if already found
			foreach ($matchesFound as $found => $set) {
				// Get test part
				$test = substr($found, 0, strlen($match));

				// Does this entry exist?
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):found={$found},match={$match},set={$set}<br />");
				if ($test == $match) {
					// Match found!
					//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):fuzzyFound!<br />");
					$fuzzyFound = true;
					break;
				} // END - if
			} // END - foreach

			// Skip this entry?
			if ($fuzzyFound === true) continue;

			// Take all string elements
			if ((is_string($matches[4][$key])) && (!isset($matchesFound[$match])) && (!isset($matchesFound[$key."_" . $matches[4][$key]]))) {
				// Replace it in the code
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):key={$key},match={$match}<br />");
				$newMatch = str_replace('[', "['", str_replace(']', "']", $match));
				$code = str_replace($match, "\"." . $newMatch.".\"", $code);
				$matchesFound[$key . '_' . $matches[4][$key]] = 1;
				$matchesFound[$match] = 1;
			} elseif (!isset($matchesFound[$match])) {
				// Not yet replaced!
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):match={$match}<br />");
				$code = str_replace($match, "\"." . $match.".\"", $code);
				$matchesFound[$match] = 1;
			}
		} // END - foreach
	} // END - if

	// Return compiled code
	return $code;
}

/************************************************************************
 *                                                                      *
 * Gaenderter Sortier-Algorythmus, $array wird nach dem Array (!)       *
 * $a_sort sortiert:                                                    *
 *                                                                      *
 * $array - Das 3-dimensionale Array, das paralell sortiert werden soll *
 * $a_sort - Array, das die Sortiereihenfolge der ersten Elementeben    *
 * $primary_key - Prim.rschl.ssel aus $a_sort, nach dem sortiert wird   *
 * $order - Sortiereihenfolge: -1 = a-Z, 0 = keine, 1 = Z-a             *
 * $nums - true = Als Zahlen sortieren, false = Als Zeichen sortieren   *
 *                                                                      *
 * $a_sort muss Elemente enthalten, deren Wert Schluessel von $array    *
 * sind... Klingt kompliziert, suchen Sie mal mein Beispiel, dann sehen *
 * Sie, dass es doch nicht so schwer ist! :-)                           *
 *                                                                      *
 ************************************************************************/
function array_pk_sort (&$array, $a_sort, $primary_key = 0, $order = -1, $nums = false) {
	$dummy = $array;
	while ($primary_key < count($a_sort)) {
		foreach ($dummy[$a_sort[$primary_key]] as $key => $value) {
			foreach ($dummy[$a_sort[$primary_key]] as $key2 => $value2) {
				$match = false;
				if ($nums === false) {
					// Sort byte-by-byte (also numbers will be interpreted as chars! E.g.: "9" > "10")
					if (($key != $key2) && (strcmp(strtolower($dummy[$a_sort[$primary_key]][$key]), strtolower($dummy[$a_sort[$primary_key]][$key2])) == $order)) $match = true;
				} elseif ($key != $key2) {
					// Sort numbers (E.g.: 9 < 10)
					if (($dummy[$a_sort[$primary_key]][$key] < $dummy[$a_sort[$primary_key]][$key2]) && ($order == -1)) $match = true;
					if (($dummy[$a_sort[$primary_key]][$key] > $dummy[$a_sort[$primary_key]][$key2]) && ($order == 1))  $match = true;
				}

				if ($match) {
					// We have found two different values, so let's sort whole array
					foreach ($dummy as $sort_key => $sort_val) {
						$t                       = $dummy[$sort_key][$key];
						$dummy[$sort_key][$key]  = $dummy[$sort_key][$key2];
						$dummy[$sort_key][$key2] = $t;
						unset($t);
					} // END - foreach
				} // END - if
			} // END - foreach
		} // END - foreach

		// Count one up
		$primary_key++;
	} // END - while

	// Write back sorted array
	$array = $dummy;
}

//
function addSelectionBox ($type, $default, $prefix = '', $id = 0) {
	$OUT = '';

	if ($type == 'yn') {
		// This is a yes/no selection only!
		if ($id > 0) $prefix .= "[" . $id."]";
		$OUT .= "    <select name=\"" . $prefix."\" class=\"register_select\" size=\"1\">\n";
	} else {
		// Begin with regular selection box here
		if (!empty($prefix)) $prefix .= "_";
		$type2 = $type;
		if ($id > 0) $type2 .= "[" . $id."]";
		$OUT .= "    <select name=\"".strtolower($prefix . $type2)."\" class=\"register_select\" size=\"1\">\n";
	}

	switch ($type) {
		case 'day': // Day
			for ($idx = 1; $idx < 32; $idx++) {
				$OUT .= "<option value=\"" . $idx."\"";
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			} // END - for
			break;

		case 'month': // Month
			foreach ($GLOBALS['month_descr'] as $month => $descr) {
				$OUT .= "<option value=\"" . $month."\"";
				if ($default == $month) $OUT .= ' selected="selected"';
				$OUT .= ">" . $descr."</option>\n";
			} // END - for
			break;

		case 'year': // Year
			// Get current year
			$year = date('Y', time());

			// Use configured min age or fixed?
			if ((isExtensionActive('other')) && (getExtensionVersion('other') >= '0.2.1')) {
				// Configured
				$startYear = $year - getConfig('min_age');
			} else {
				// Fixed 16 years
				$startYear = $year - 16;
			}

			// Calculate earliest year (100 years old people can still enter Internet???)
			$minYear = $year - 100;

			// Check if the default value is larger than minimum and bigger than actual year
			if (($default > $minYear) && ($default >= $year)) {
				for ($idx = $year; $idx < ($year + 11); $idx++) {
					$OUT .= "<option value=\"" . $idx."\"";
					if ($default == $idx) $OUT .= ' selected="selected"';
					$OUT .= ">" . $idx."</option>\n";
				} // END - for
			} elseif ($default == -1) {
				// Current year minus 1
				for ($idx = $startYear; $idx <= ($year + 1); $idx++)
				{
					$OUT .= "<option value=\"" . $idx."\">" . $idx."</option>\n";
				}
			} else {
				// Get current year and subtract the configured minimum age
				$OUT .= "<option value=\"".($minYear - 1)."\">&lt;" . $minYear."</option>\n";
				// Calculate earliest year depending on extension version
				if ((isExtensionActive('other')) && (getExtensionVersion('other') >= '0.2.1')) {
					// Use configured minimum age
					$year = date('Y', time()) - getConfig('min_age');
				} else {
					// Use fixed 16 years age
					$year = date('Y', time()) - 16;
				}

				// Construct year selection list
				for ($idx = $minYear; $idx <= $year; $idx++) {
					$OUT .= "<option value=\"" . $idx."\"";
					if ($default == $idx) $OUT .= ' selected="selected"';
					$OUT .= ">" . $idx."</option>\n";
				} // END - for
			}
			break;

		case 'sec':
		case 'min':
			for ($idx = 0; $idx < 60; $idx+=5) {
				if (strlen($idx) == 1) $idx = 0 . $idx;
				$OUT .= "<option value=\"" . $idx."\"";
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			} // END - for
			break;

		case 'hour':
			for ($idx = 0; $idx < 24; $idx++) {
				if (strlen($idx) == 1) $idx = 0 . $idx;
				$OUT .= "<option value=\"" . $idx."\"";
				if ($default == $idx) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			} // END - for
			break;

		case 'yn':
			$OUT .= "<option value=\"Y\"";
			if ($default == 'Y') $OUT .= ' selected="selected"';
			$OUT .= ">{--YES--}</option>\n<option value=\"N\"";
			if ($default != 'Y') $OUT .= ' selected="selected"';
			$OUT .= ">{--NO--}</option>\n";
			break;
	}
	$OUT .= "    </select>\n";
	return $OUT;
}

//
// Deprecated : $length
// Optional   : $DATA
//
function generateRandomCode ($length, $code, $userid, $DATA = '') {
	// Build server string
	$server = $_SERVER['PHP_SELF'] . getConfig('ENCRYPT_SEPERATOR') . detectUserAgent() . getConfig('ENCRYPT_SEPERATOR') . getenv('SERVER_SOFTWARE') . getConfig('ENCRYPT_SEPERATOR') . detectRemoteAddr().":'.':".filemtime(getConfig('PATH').'inc/databases.php');

	// Build key string
	$keys = getConfig('SITE_KEY') . getConfig('ENCRYPT_SEPERATOR') . getConfig('DATE_KEY');
	if (isConfigEntrySet('secret_key'))  $keys .= getConfig('ENCRYPT_SEPERATOR').getConfig('secret_key');
	if (isConfigEntrySet('file_hash'))   $keys .= getConfig('ENCRYPT_SEPERATOR').getConfig('file_hash');
	$keys .= getConfig('ENCRYPT_SEPERATOR') . date('d-m-Y (l-F-T)', getConfig('patch_ctime'));
	if (isConfigEntrySet('master_salt')) $keys .= getConfig('ENCRYPT_SEPERATOR').getConfig('master_salt');

	// Build string from misc data
	$data   = $code . getConfig('ENCRYPT_SEPERATOR') . $userid . getConfig('ENCRYPT_SEPERATOR') . $DATA;

	// Add more additional data
	if (isSessionVariableSet('u_hash'))         $data .= getConfig('ENCRYPT_SEPERATOR') . getSession('u_hash');

	// Add referal id, language, theme and userid
	$data .= getConfig('ENCRYPT_SEPERATOR') . determineReferalId();
	$data .= getConfig('ENCRYPT_SEPERATOR') . getLanguage();
	$data .= getConfig('ENCRYPT_SEPERATOR') . getCurrentTheme();
	$data .= getConfig('ENCRYPT_SEPERATOR') . getUserId();

	// Calculate number for generating the code
	$a = $code + getConfig('_ADD') - 1;

	if (isConfigEntrySet('master_hash')) {
		// Generate hash with master salt from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % getConfig('_PRIME')) . getConfig('ENCRYPT_SEPERATOR') . $server . getConfig('ENCRYPT_SEPERATOR') . $keys . getConfig('ENCRYPT_SEPERATOR') . $data . getConfig('ENCRYPT_SEPERATOR') . getConfig('DATE_KEY') . getConfig('ENCRYPT_SEPERATOR') . $a, getConfig('master_salt'));

		// Create number from hash
		$rcode = hexdec(substr($saltedHash, strlen(getConfig('master_salt')), 9)) / abs(getConfig('rand_no') - $a + sqrt(getConfig('_ADD'))) / pi();
	} else {
		// Generate hash with "hash of site key" from modula of number with the prime number and other data
		$saltedHash = generateHash(($a % getConfig('_PRIME')) . getConfig('ENCRYPT_SEPERATOR') . $server . getConfig('ENCRYPT_SEPERATOR') . $keys . getConfig('ENCRYPT_SEPERATOR') . $data . getConfig('ENCRYPT_SEPERATOR') . getConfig('DATE_KEY') . getConfig('ENCRYPT_SEPERATOR') . $a, substr(sha1(getConfig('SITE_KEY')), 0, 8));

		// Create number from hash
		$rcode = hexdec(substr($saltedHash, 8, 9)) / abs(getConfig('rand_no') - $a + sqrt(getConfig('_ADD'))) / pi();
	}

	// At least 10 numbers shall be secure enought!
	$len = getConfig('code_length');
	if ($len == 0) $len = $length;
	if ($len == 0) $len = 10;

	// Cut off requested counts of number
	$return = substr(str_replace('.', '', $rcode), 0, $len);

	// Done building code
	return $return;
}

// Does only allow numbers
function bigintval ($num, $castValue = true) {
	// Filter all numbers out
	$ret = preg_replace('/[^0123456789]/', '', $num);

	// Shall we cast?
	if ($castValue) $ret = (double)$ret;

	// Has the whole value changed?
	// @TODO Remove this if() block if all is working fine
	if ('' . $ret . '' != '' . $num . '') {
		// Log the values
		//debug_report_bug("{$ret}<>{$num}");
	} // END - if

	// Return result
	return $ret;
}

// Insert the code in $img_code into jpeg or PNG image
function generateImageOrCode ($img_code, $headerSent=true) {
	if ((strlen($img_code) > 6) || (empty($img_code)) || (getConfig('code_length') == 0)) {
		// Stop execution of function here because of over-sized code length
		return;
	} elseif ($headerSent === false) {
		// Return in an HTML code code
		return "<img src=\"{?URL?}/img.php?code=" . $img_code."\" alt=\"Image\" />\n";
	}

	// Load image
	$img = sprintf("%s/theme/%s/images/code_bg.%s", getConfig('PATH'), getCurrentTheme(), getConfig('img_type'));
	if (isFileReadable($img)) {
		// Switch image type
		switch (getConfig('img_type'))
		{
			case 'jpg':
				// Okay, load image and hide all errors
				$image = imagecreatefromjpeg($img);
				break;

			case 'png':
				// Okay, load image and hide all errors
				$image = imagecreatefrompng($img);
				break;
		}
	} else {
		// Exit function here
		logDebugMessage(__FUNCTION__, __LINE__, sprintf("File for image type %s not found.", getConfig('img_type')));
		return;
	}

	// Generate text color (red/green/blue; 0 = dark, 255 = bright)
	$text_color = imagecolorallocate($image, 0, 0, 0);

	// Insert code into image
	imagestring($image, 5, 14, 2, $img_code, $text_color);

	// Return to browser
	sendHeader('Content-Type: image/' . getConfig('img_type'));

	// Output image with matching image factory
	switch (getConfig('img_type')) {
		case 'jpg': imagejpeg($image); break;
		case 'png': imagepng($image);  break;
	}

	// Remove image from memory
	imagedestroy($image);
}
// Create selection box or array of splitted timestamp
function createTimeSelections ($timestamp, $prefix = '', $display = '', $align = 'center', $return_array=false) {
	// Calculate 2-seconds timestamp
	$stamp = round($timestamp);
	//* DEBUG: */ print("*" . $stamp.'/' . $timestamp."*<br />");

	// Do we have a leap year?
	$SWITCH = 0;
	$TEST = date('Y', time()) / 4;
	$M1 = date('m', time());
	$M2 = date('m', (time() + $timestamp));

	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == "02") && ($M2 > "02"))  $SWITCH = getConfig('ONE_DAY');

	// First of all years...
	$Y = abs(floor($timestamp / (31536000 + $SWITCH)));
	//* DEBUG: */ print("Y={$Y}<br />");
	// Next months...
	$M = abs(floor($timestamp / 2628000 - $Y * 12));
	//* DEBUG: */ print("M={$M}<br />");
	// Next weeks
	$W = abs(floor($timestamp / 604800 - $Y * ((365 + $SWITCH / getConfig('ONE_DAY')) / 7) - ($M / 12 * (365 + $SWITCH / getConfig('ONE_DAY')) / 7)));
	//* DEBUG: */ print("W={$W}<br />");
	// Next days...
	$D = abs(floor($timestamp / 86400 - $Y * (365 + $SWITCH / getConfig('ONE_DAY')) - ($M / 12 * (365 + $SWITCH / getConfig('ONE_DAY'))) - $W * 7));
	//* DEBUG: */ print("D={$D}<br />");
	// Next hours...
	$h = abs(floor($timestamp / 3600 - $Y * (365 + $SWITCH / getConfig('ONE_DAY')) * 24 - ($M / 12 * (365 + $SWITCH / getConfig('ONE_DAY')) * 24) - $W * 7 * 24 - $D * 24));
	//* DEBUG: */ print("h={$h}<br />");
	// Next minutes..
	$m = abs(floor($timestamp / 60 - $Y * (365 + $SWITCH / getConfig('ONE_DAY')) * 24 * 60 - ($M / 12 * (365 + $SWITCH / getConfig('ONE_DAY')) * 24 * 60) - $W * 7 * 24 * 60 - $D * 24 * 60 - $h * 60));
	//* DEBUG: */ print("m={$m}<br />");
	// And at last seconds...
	$s = abs(floor($timestamp - $Y * (365 + $SWITCH / getConfig('ONE_DAY')) * 24 * 3600 - ($M / 12 * (365 + $SWITCH / getConfig('ONE_DAY')) * 24 * 3600) - $W * 7 * 24 * 3600 - $D * 24 * 3600 - $h * 3600 - $m * 60));
	//* DEBUG: */ print("s={$s}<br />");

	// Is seconds zero and time is < 60 seconds?
	if (($s == 0) && ($timestamp < 60)) {
		// Fix seconds
		$s = round($timestamp);
	} // END - if

	//
	// Now we convert them in seconds...
	//
	if ($return_array) {
		// Just put all data in an array for later use
		$OUT = array(
			'YEARS'   => $Y,
			'MONTHS'  => $M,
			'WEEKS'   => $W,
			'DAYS'    => $D,
			'HOURS'   => $h,
			'MINUTES' => $m,
			'SECONDS' => $s
		);
	} else {
		// Generate table
		$OUT  = "<div align=\"" . $align."\">\n";
		$OUT .= "<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\" class=\"admin_table dashed\">\n";
		$OUT .= "<tr>\n";

		if (ereg('Y', $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_YEARS--}</strong></td>\n";
		}

		if (ereg('M', $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_MONTHS--}</strong></td>\n";
		}

		if (ereg("W", $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_WEEKS--}</strong></td>\n";
		}

		if (ereg("D", $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_DAYS--}</strong></td>\n";
		}

		if (ereg("h", $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_HOURS--}</strong></td>\n";
		}

		if (ereg('m', $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_MINUTES--}</strong></td>\n";
		}

		if (ereg("s", $display) || (empty($display))) {
			$OUT .= "  <td align=\"center\" class=\"admin_title bottom\"><div class=\"tiny\">{--_SECONDS--}</strong></td>\n";
		}

		$OUT .= "</tr>\n";
		$OUT .= "<tr>\n";

		if (ereg('Y', $display) || (empty($display))) {
			// Generate year selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_ye\" size=\"1\">\n";
			for ($idx = 0; $idx <= 10; $idx++) {
				$OUT .= "    <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $Y) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_ye\" value=\"0\" />\n";
		}

		if (ereg('M', $display) || (empty($display))) {
			// Generate month selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_mo\" size=\"1\">\n";
			for ($idx = 0; $idx <= 11; $idx++)
			{
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $M) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_mo\" value=\"0\" />\n";
		}

		if (ereg("W", $display) || (empty($display))) {
			// Generate week selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_we\" size=\"1\">\n";
			for ($idx = 0; $idx <= 4; $idx++) {
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $W) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_we\" value=\"0\" />\n";
		}

		if (ereg("D", $display) || (empty($display))) {
			// Generate day selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_da\" size=\"1\">\n";
			for ($idx = 0; $idx <= 31; $idx++) {
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $D) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_da\" value=\"0\">\n";
		}

		if (ereg("h", $display) || (empty($display))) {
			// Generate hour selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_ho\" size=\"1\">\n";
			for ($idx = 0; $idx <= 23; $idx++) 	{
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $h) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_ho\" value=\"0\">\n";
		}

		if (ereg('m', $display) || (empty($display))) {
			// Generate minute selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_mi\" size=\"1\">\n";
			for ($idx = 0; $idx <= 59; $idx++) {
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $m) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_mi\" value=\"0\">\n";
		}

		if (ereg("s", $display) || (empty($display))) {
			// Generate second selection
			$OUT .= "  <td align=\"center\"><select class=\"mini_select\" name=\"" . $prefix."_se\" size=\"1\">\n";
			for ($idx = 0; $idx <= 59; $idx++) {
				$OUT .= "  <option class=\"mini_select\" value=\"" . $idx."\"";
				if ($idx == $s) $OUT .= ' selected="selected"';
				$OUT .= ">" . $idx."</option>\n";
			}
			$OUT .= "  </select></td>\n";
		} else {
			$OUT .= "<INPUT type=\"hidden\" name=\"" . $prefix."_se\" value=\"0\">\n";
		}
		$OUT .= "</tr>\n";
		$OUT .= "</table>\n";
		$OUT .= "</div>\n";
		// Return generated HTML code
	}
	return $OUT;
}

//
function createTimestampFromSelections ($prefix, $postData) {
	// Initial return value
	$ret = 0;

	// Do we have a leap year?
	$SWITCH = 0;
	$TEST = date('Y', time()) / 4;
	$M1   = date('m', time());
	// If so and if current time is before 02/29 and estimated time is after 02/29 then add 86400 seconds (one day)
	if ((floor($TEST) == $TEST) && ($M1 == "02") && ($postData[$prefix."_mo"] > "02"))  $SWITCH = getConfig('ONE_DAY');
	// First add years...
	$ret += $postData[$prefix."_ye"] * (31536000 + $SWITCH);
	// Next months...
	$ret += $postData[$prefix."_mo"] * 2628000;
	// Next weeks
	$ret += $postData[$prefix."_we"] * 604800;
	// Next days...
	$ret += $postData[$prefix."_da"] * 86400;
	// Next hours...
	$ret += $postData[$prefix."_ho"] * 3600;
	// Next minutes..
	$ret += $postData[$prefix."_mi"] * 60;
	// And at last seconds...
	$ret += $postData[$prefix."_se"];
	// Return calculated value
	return $ret;
}

// Creates a 'fancy' human-readable timestamp from a Uni* stamp
function createFancyTime ($stamp) {
	// Get data array with years/months/weeks/days/...
	$data = createTimeSelections($stamp, '', '', '', true);
	$ret = '';
	foreach($data as $k => $v) {
		if ($v > 0) {
			// Value is greater than 0 "eval" data to return string
			eval("\$ret .= \", \".\$v.\" {--_".strtoupper($k)."--}\";");
			break;
		} // END - if
	} // END - foreach

	// Do we have something there?
	if (strlen($ret) > 0) {
		// Remove leading commata and space
		$ret = substr($ret, 2);
	} else {
		// Zero seconds
		$ret = "0 {--_SECONDS--}";
	}

	// Return fancy time string
	return $ret;
}

// Generates a navigation row for listing emails
function addEmailNavigation ($PAGES, $offset, $show_form, $colspan, $return=false) {
	$SEP = ''; $TOP = '';
	if ($show_form === false) {
		$TOP = " top2";
		$SEP = "<tr><td colspan=\"" . $colspan."\" class=\"seperator\">&nbsp;</td></tr>";
	}

	$NAV = '';
	for ($page = 1; $page <= $PAGES; $page++) {
		// Is the page currently selected or shall we generate a link to it?
		if (($page == getRequestElement('page')) || ((!isGetRequestElementSet('page')) && ($page == 1))) {
			// Is currently selected, so only highlight it
			$NAV .= '<strong>-';
		} else {
			// Open anchor tag and add base URL
			$NAV .= '<a href="{?URL?}/modules.php?module=admin&amp;what=' . getWhat() . '&amp;page=' . $page . '&amp;offset=' . $offset;

			// Add userid when we shall show all mails from a single member
			if ((isGetRequestElementSet('userid')) && (bigintval(getRequestElement('userid')) > 0)) $NAV .= '&amp;userid=' . bigintval(getRequestElement('userid'));

			// Close open anchor tag
			$NAV .= '">';
		}
		$NAV .= $page;
		if (($page == getRequestElement('page')) || ((!isGetRequestElementSet('page')) && ($page == 1))) {
			// Is currently selected, so only highlight it
			$NAV .= '-</strong>';
		} else {
			// Close anchor tag
			$NAV .= '</a>';
		}

		// Add seperator if we have not yet reached total pages
		if ($page < $PAGES) $NAV .= '&nbsp;|&nbsp;';
	} // END - for

	// Define constants only once
	$content['nav']  = $NAV;
	$content['span'] = $colspan;
	$content['top']  = $TOP;
	$content['sep']  = $SEP;

	// Load navigation template
	$OUT = loadTemplate('admin_email_nav_row', true, $content);

	if ($return === true) {
		// Return generated HTML-Code
		return $OUT;
	} else {
		// Output HTML-Code
		outputHtml($OUT);
	}
}

// Extract host from script name
function extractHostnameFromUrl (&$script) {
	// Use default SERVER_URL by default... ;) So?
	$url = getConfig('SERVER_URL');

	// Is this URL valid?
	if (substr($script, 0, 7) == 'http://') {
		// Use the hostname from script URL as new hostname
		$url = substr($script, 7);
		$extract = explode('/', $url);
		$url = $extract[0];
		// Done extracting the URL :)
	} // END - if

	// Extract host name
	$host = str_replace('http://', '', $url);
	if (ereg('/', $host)) $host = substr($host, 0, strpos($host, '/'));

	// Generate relative URL
	//* DEBUG: */ print("SCRIPT=" . $script."<br />");
	if (substr(strtolower($script), 0, 7) == 'http://') {
		// But only if http:// is in front!
		$script = substr($script, (strlen($url) + 7));
	} elseif (substr(strtolower($script), 0, 8) == "https://") {
		// Does this work?!
		$script = substr($script, (strlen($url) + 8));
	}

	//* DEBUG: */ print("SCRIPT=" . $script."<br />");
	if (substr($script, 0, 1) == '/') $script = substr($script, 1);

	// Return host name
	return $host;
}

// Send a GET request
function sendGetRequest ($script, $data = array()) {
	// Extract host name from script
	$host = extractHostnameFromUrl($script);

	// Add data
	$scriptData = http_build_query($data, '', '&');

	// Do we have a question-mark in the script?
	if (strpos($script, '?') === false) {
		// No, so first char must be question mark
		$scriptData = '?' . $scriptData;
	} else {
		// Ok, add &
		$scriptData = '&' . $scriptData;
	}

	// Add script data
	$script .= $scriptData;

	// Generate GET request header
	$request  = 'GET /' . trim($script) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getConfig('URL') . '/admin.php' . getConfig('HTTP_EOL');
	if (isConfigEntrySet('FULL_VERSION')) {
		$request .= 'User-Agent: ' . getConfig('TITLE') . '/' . getConfig('FULL_VERSION') . getConfig('HTTP_EOL');
	} else {
		$request .= 'User-Agent: ' . getConfig('TITLE') . '/' . getConfig('VERSION') . getConfig('HTTP_EOL');
	}
	$request .= 'Content-Type: text/plain' . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-cache' . getConfig('HTTP_EOL');
	$request .= 'Connection: Close' . getConfig('HTTP_EOL') . getConfig('HTTP_EOL');

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Return the result to the caller function
	return $response;
}

// Send a POST request
function sendPostRequest ($script, $postData) {
	// Is postData an array?
	if (!is_array($postData)) {
		// Abort here
		logDebugMessage(__FUNCTION__, __LINE__, sprintf("postData is not an array. Type: %s", gettype($postData)));
		return array('', '', '');
	} // END - if

	// Compile the script name
	$script = compileCode($script);

	// Extract host name from script
	$host = extractHostnameFromUrl($script);

	// Construct request
	$data = http_build_query($postData, '', '&');

	// Generate POST request header
	$request  = 'POST /' . trim($script) . ' HTTP/1.1' . getConfig('HTTP_EOL');
	$request .= 'Host: ' . $host . getConfig('HTTP_EOL');
	$request .= 'Referer: ' . getConfig('URL') . '/admin.php' . getConfig('HTTP_EOL');
	$request .= 'User-Agent: ' . getConfig('TITLE') . '/' . getConfig('FULL_VERSION') . getConfig('HTTP_EOL');
	$request .= 'Content-Type: application/x-www-form-urlencoded' . getConfig('HTTP_EOL');
	$request .= 'Content-length: ' . strlen($data) . getConfig('HTTP_EOL');
	$request .= 'Cache-Control: no-cache' . getConfig('HTTP_EOL');
	$request .= 'Connection: Close' . getConfig('HTTP_EOL') . getConfig('HTTP_EOL');
	$request .= $data;

	// Send the raw request
	$response = sendRawRequest($host, $request);

	// Return the result to the caller function
	return $response;
}

// Sends a raw request to another host
function sendRawRequest ($host, $request) {
	// Init errno and errdesc with 'all fine' values
	$errno = 0; $errdesc = '';

	// Initialize array
	$response = array('', '', '');

	// Default is not to use proxy
	$useProxy = false;

	// Are proxy settins set?
	if ((isConfigEntrySet('proxy_host')) && (getConfig('proxy_host') != '') && (isConfigEntrySet('proxy_port')) && (getConfig('proxy_port') > 0)) {
		// Then use it
		$useProxy = true;
	} // END - if

	// Open connection
	//* DEBUG: */ die("SCRIPT=" . $script."<br />");
	if ($useProxy === true) {
		// Connect to host through proxy connection
		$fp = @fsockopen(compileCode(getConfig('proxy_host')), bigintval(getConfig('proxy_port')), $errno, $errdesc, 30);
	} else {
		// Connect to host directly
		$fp = @fsockopen($host, 80, $errno, $errdesc, 30);
	}

	// Is there a link?
	if (!is_resource($fp)) {
		// Failed!
		return $response;
	} // END - if

	// Do we use proxy?
	if ($useProxy === true) {
		// Generate CONNECT request header
		$proxyTunnel  = "CONNECT " . $host . ":80 HTTP/1.1" . getConfig('HTTP_EOL');
		$proxyTunnel .= "Host: " . $host . getConfig('HTTP_EOL');

		// Use login data to proxy? (username at least!)
		if (getConfig('proxy_username') != '') {
			// Add it as well
			$encodedAuth = base64_encode(compileCode(getConfig('proxy_username')) . getConfig('ENCRYPT_SEPERATOR') . compileCode(getConfig('proxy_password')));
			$proxyTunnel .= "Proxy-Authorization: Basic " . $encodedAuth . getConfig('HTTP_EOL');
		} // END - if

		// Add last new-line
		$proxyTunnel .= getConfig('HTTP_EOL');
		//* DEBUG: */ print("<strong>proxyTunnel=</strong><pre>" . $proxyTunnel."</pre>");

		// Write request
		fputs($fp, $proxyTunnel);

		// Got response?
		if (feof($fp)) {
			// No response received
			return $response;
		} // END - if

		// Read the first line
		$resp = trim(fgets($fp, 10240));
		$respArray = explode(' ', $resp);
		if ((strtolower($respArray[0]) !== 'http/1.0') || ($respArray[1] != '200')) {
			// Invalid response!
			return $response;
		} // END - if
	} // END - if

	// Write request
	fputs($fp, $request);

	// Read response
	while (!feof($fp)) {
		$response[] = trim(fgets($fp, 1024));
	} // END - while

	// Close socket
	fclose($fp);

	// Skip first empty lines
	$resp = $response;
	foreach ($resp as $idx => $line) {
		// Trim space away
		$line = trim($line);

		// Is this line empty?
		if (empty($line)) {
			// Then remove it
			array_shift($response);
		} else {
			// Abort on first non-empty line
			break;
		}
	} // END - foreach

	//* DEBUG: */ print("<strong>Response:</strong><pre>".print_r($response, true)."</pre>");

	// Proxy agent found?
	if ((substr(strtolower($response[0]), 0, 11) == 'proxy-agent') && ($useProxy === true)) {
		// Proxy header detected, so remove two lines
		array_shift($response);
		array_shift($response);
	} // END - if

	// Was the request successfull?
	if ((!eregi('200 OK', $response[0])) || (empty($response[0]))) {
		// Not found / access forbidden
		$response = array('', '', '');
	} // END - if

	// Return response
	return $response;
}

// Taken from www.php.net eregi() user comments
function isEmailValid ($email) {
	// Compile email
	$email = compileCode($email);

	// Check first part of email address
	$first = '[-a-z0-9!#$%&\'*+/=?^_<{|}~]+(\.[-a-zA-Z0-9!#$%&\'*+/=?^_<{|}~]+)*';

	//  Check domain
	$domain = '[a-z0-9-]+(\.[a-z0-9-]{2,5})+';

	// Generate pattern
	$regex = '@^' . $first . '\@' . $domain . '$@iU';

	// Return check result
	return preg_match($regex, $email);
}

// Function taken from user comments on www.php.net / function eregi()
function isUrlValid ($URL, $compile=true) {
	// Trim URL a little
	$URL = trim(urldecode($URL));
	//* DEBUG: */ outputHtml($URL."<br />");

	// Compile some chars out...
	if ($compile === true) $URL = compileUriCode($URL, false, false, false);
	//* DEBUG: */ outputHtml($URL."<br />");

	// Check for the extension filter
	if (isExtensionActive('filter')) {
		// Use the extension's filter set
		return FILTER_VALIDATE_URL($URL, false);
	} // END - if

	// If not installed, perform a simple test. Just make it sure there is always a http:// or
	// https:// in front of the URLs
	return isUrlValidSimple($URL);
}

// Generate a list of administrative links to a given userid
function generateMemberAdminActionLinks ($userid, $status = '') {
	// Make sure userid is a number
	if ($userid != bigintval($userid)) debug_report_bug('userid is not a number!');

	// Define all main targets
	$targetArray = array('del_user', 'edit_user', 'lock_user', 'add_points', 'sub_points');

	// Begin of navigation links
	$OUT = "[&nbsp;";

	foreach ($targetArray as $tar) {
		$OUT .= "<span class=\"admin_user_link\"><a href=\"{?URL?}/modules.php?module=admin&amp;what=" . $tar . "&amp;userid=" . $userid . "\" title=\"{--ADMIN_LINK_";
		//* DEBUG: */ outputHtml("*" . $tar.'/' . $status."*<br />");
		if (($tar == 'lock_user') && ($status == 'LOCKED')) {
			// Locked accounts shall be unlocked
			$OUT .= 'UNLOCK_USER';
		} else {
			// All other status is fine
			$OUT .= strtoupper($tar);
		}
		$OUT .= "_TITLE--}\">{--ADMIN_";
		if (($tar == 'lock_user') && ($status == 'LOCKED')) {
			// Locked accounts shall be unlocked
			$OUT .= 'UNLOCK_USER';
		} else {
			// All other status is fine
			$OUT .= strtoupper($tar);
		}
		$OUT .= "--}</a></span>&nbsp;|&nbsp;";
	}

	// Finish navigation link
	$OUT = substr($OUT, 0, -7) . ']';

	// Return string
	return $OUT;
}

// Generate an email link
function generateEmailLink ($email, $table = 'admins') {
	// Default email link (INSECURE! Spammer can read this by harvester programs)
	$EMAIL = 'mailto:' . $email;

	// Check for several extensions
	if ((isExtensionActive('admins')) && ($table == 'admins')) {
		// Create email link for contacting admin in guest area
		$EMAIL = generateAdminEmailLink($email);
	} elseif ((isExtensionActive('user')) && (getExtensionVersion('user') >= '0.3.3') && ($table == 'user_data')) {
		// Create email link for contacting a member within admin area (or later in other areas, too?)
		$EMAIL = generateUserEmailLink($email, 'admin');
	} elseif ((isExtensionActive('sponsor')) && ($table == 'sponsor_data')) {
		// Create email link to contact sponsor within admin area (or like the link above?)
		$EMAIL = generateSponsorEmailLink($email, 'sponsor_data');
	}

	// Shall I close the link when there is no admin?
	if ((!isAdmin()) && ($EMAIL == $email)) $EMAIL = '#'; // Closed!

	// Return email link
	return $EMAIL;
}

// Generate a hash for extra-security for all passwords
function generateHash ($plainText, $salt = '') {
	// Is the required extension 'sql_patches' there and a salt is not given?
	if (((isExtensionInstalledAndOlder('sql_patches', '0.3.6')) || (!isExtensionActive('sql_patches'))) && (empty($salt))) {
		// Extension sql_patches is missing/outdated so we hash the plain text with MD5
		return md5($plainText);
	} // END - if

	// Do we miss an arry element here?
	if (!isConfigEntrySet('file_hash')) {
		// Stop here
		debug_report_bug('Missing file_hash in ' . __FUNCTION__ . '.');
	} // END - if

	// When the salt is empty build a new one, else use the first x configured characters as the salt
	if (empty($salt)) {
		// Build server string (inc/databases.php is no longer updated with every commit)
		$server = $_SERVER['PHP_SELF'] . getConfig('ENCRYPT_SEPERATOR') . detectUserAgent() . getConfig('ENCRYPT_SEPERATOR') . getenv('SERVER_SOFTWARE') . getConfig('ENCRYPT_SEPERATOR') . detectRemoteAddr();

		// Build key string
		$keys   = getConfig('SITE_KEY') . getConfig('ENCRYPT_SEPERATOR') . getConfig('DATE_KEY') . getConfig('ENCRYPT_SEPERATOR') . getConfig('secret_key') . getConfig('ENCRYPT_SEPERATOR') . getConfig('file_hash') . getConfig('ENCRYPT_SEPERATOR') . date('d-m-Y (l-F-T)', getConfig('patch_ctime')) . getConfig('ENCRYPT_SEPERATOR') . getConfig('master_salt');

		// Additional data
		$data = $plainText . getConfig('ENCRYPT_SEPERATOR') . uniqid(mt_rand(), true) . getConfig('ENCRYPT_SEPERATOR') . time();

		// Calculate number for generating the code
		$a = time() + getConfig('_ADD') - 1;

		// Generate SHA1 sum from modula of number and the prime number
		$sha1 = sha1(($a % getConfig('_PRIME')) . $server . getConfig('ENCRYPT_SEPERATOR') . $keys . getConfig('ENCRYPT_SEPERATOR') . $data . getConfig('ENCRYPT_SEPERATOR') . getConfig('DATE_KEY') . getConfig('ENCRYPT_SEPERATOR') . $a);
		//* DEBUG: */ outputHtml("SHA1=" . $sha1." (".strlen($sha1).")<br />");
		$sha1 = scrambleString($sha1);
		//* DEBUG: */ outputHtml("Scrambled=" . $sha1." (".strlen($sha1).")<br />");
		//* DEBUG: */ $sha1b = descrambleString($sha1);
		//* DEBUG: */ outputHtml("Descrambled=" . $sha1b." (".strlen($sha1b).")<br />");

		// Generate the password salt string
		$salt = substr($sha1, 0, getConfig('salt_length'));
		//* DEBUG: */ outputHtml($salt." (".strlen($salt).")<br />");
	} else {
		// Use given salt
		$salt = substr($salt, 0, getConfig('salt_length'));
		//* DEBUG: */ outputHtml("GIVEN={$salt}<br />");
	}

	// Return hash
	return $salt.sha1($salt . $plainText);
}

// Scramble a string
function scrambleString($str) {
	// Init
	$scrambled = '';

	// Final check, in case of failture it will return unscrambled string
	if (strlen($str) > 40) {
		// The string is to long
		return $str;
	} elseif (strlen($str) == 40) {
		// From database
		$scrambleNums = explode(':', getConfig('pass_scramble'));
	} else {
		// Generate new numbers
		$scrambleNums = explode(':', genScrambleString(strlen($str)));
	}

	// Scramble string here
	//* DEBUG: */ outputHtml("***Original=" . $str."***<br />");
	for ($idx = 0; $idx < strlen($str); $idx++) {
		// Get char on scrambled position
		$char = substr($str, $scrambleNums[$idx], 1);

		// Add it to final output string
		$scrambled .= $char;
	} // END - for

	// Return scrambled string
	//* DEBUG: */ outputHtml("***Scrambled=" . $scrambled."***<br />");
	return $scrambled;
}

// De-scramble a string scrambled by scrambleString()
function descrambleString($str) {
	// Scramble only 40 chars long strings
	if (strlen($str) != 40) return $str;

	// Load numbers from config
	$scrambleNums = explode(':', getConfig('pass_scramble'));

	// Validate numbers
	if (count($scrambleNums) != 40) return $str;

	// Begin descrambling
	$orig = str_repeat(' ', 40);
	//* DEBUG: */ outputHtml("+++Scrambled=" . $str."+++<br />");
	for ($idx = 0; $idx < 40; $idx++) {
		$char = substr($str, $idx, 1);
		$orig = substr_replace($orig, $char, $scrambleNums[$idx], 1);
	} // END - for

	// Return scrambled string
	//* DEBUG: */ outputHtml("+++Original=" . $orig."+++<br />");
	return $orig;
}

// Generated a "string" for scrambling
function genScrambleString ($len) {
	// Prepare array for the numbers
	$scrambleNumbers = array();

	// First we need to setup randomized numbers from 0 to 31
	for ($idx = 0; $idx < $len; $idx++) {
		// Generate number
		$rand = mt_rand(0, ($len -1));

		// Check for it by creating more numbers
		while (array_key_exists($rand, $scrambleNumbers)) {
			$rand = mt_rand(0, ($len -1));
		} // END - while

		// Add number
		$scrambleNumbers[$rand] = $rand;
	} // END - for

	// So let's create the string for storing it in database
	$scrambleString = implode(':', $scrambleNumbers);
	return $scrambleString;
}

// Generate an PGP-like encrypted hash of given hash for e.g. cookies
function generatePassString ($passHash) {
	// Return vanilla password hash
	$ret = $passHash;

	// Is a secret key and master salt already initialized?
	if ((isExtensionInstalled('sql_patches')) && (isConfigEntrySet('secret_key')) && (isConfigEntrySet('master_salt'))) {
		// Only calculate when the secret key is generated
		$newHash = ''; $start = 9;
		for ($idx = 0; $idx < 10; $idx++) {
			$part1 = hexdec(substr($passHash, $start, 4));
			$part2 = hexdec(substr(getConfig('secret_key'), $start, 4));
			$mod = dechex($idx);
			if ($part1 > $part2) {
				$mod = dechex(sqrt(($part1 - $part2) * getConfig('_PRIME') / pi()));
			} elseif ($part2 > $part1) {
				$mod = dechex(sqrt(($part2 - $part1) * getConfig('_PRIME') / pi()));
			}
			$mod = substr(round($mod), 0, 4);
			$mod = str_repeat(0, 4-strlen($mod)) . $mod;
			//* DEBUG: */ outputHtml("*" . $start.'=' . $mod."*<br />");
			$start += 4;
			$newHash .= $mod;
		} // END - for

		//* DEBUG: */ print($passHash."<br />" . $newHash." (".strlen($newHash).')');
		$ret = generateHash($newHash, getConfig('master_salt'));
		//* DEBUG: */ print($ret."<br />");
	} else {
		// Hash it simple
		//* DEBUG: */ outputHtml("--" . $passHash."--<br />");
		$ret = md5($passHash);
		//* DEBUG: */ outputHtml("++" . $ret."++<br />");
	}

	// Return result
	return $ret;
}

// Fix "deleted" cookies
function fixDeletedCookies ($cookies) {
	// Is this an array with entries?
	if ((is_array($cookies)) && (count($cookies) > 0)) {
		// Then check all cookies if they are marked as deleted!
		foreach ($cookies as $cookieName) {
			// Is the cookie set to "deleted"?
			if (getSession($cookieName) == 'deleted') {
				setSession($cookieName, '');
			} // END - if
		} // END - foreach
	} // END - if
}

// Output error messages in a fasioned way and die...
function app_die ($F, $L, $message) {
	// Check if Script is already dieing and not let it kill itself another 1000 times
	if (!isset($GLOBALS['app_died'])) {
		// Make sure, that the script realy realy diese here and now
		$GLOBALS['app_died'] = true;

		// Load header
		loadIncludeOnce('inc/header.php');

		// Rewrite message for output
		$message = sprintf(getMessage('MXCHANGE_HAS_DIED'), basename($F), $L, $message);

		// Better log this message away
		logDebugMessage($F, $L, $message);

		// Load the message template
		loadTemplate('admin_settings_saved', false, $message);

		// Load footer
		loadIncludeOnce('inc/footer.php');
	} else {
		// Script tried to kill itself twice
		debug_report_bug('Script wanted to kill itself more than once! Raw message=' . $message . ', file/function=' . $F . ', line=' . $L);
	}
}

// Display parsing time and number of SQL queries in footer
function displayParsingTime() {
	// Is the timer started?
	if (!isset($GLOBALS['startTime'])) {
		// Abort here
		return false;
	} // END - if

	// Get end time
	$endTime = microtime(true);

	// "Explode" both times
	$start = explode(' ', $GLOBALS['startTime']);
	$end = explode(' ', $endTime);
	$runTime = $end[0] - $start[0];
	if ($runTime < 0) $runTime = 0;

	// Prepare output
	$content = array(
		'runtime'  => translateComma($runTime),
		'timeSQLs' => translateComma(getConfig('sql_time') * 1000),
	);

	// Load the template
	loadTemplate('show_timings', false, $content);
}

// Check wether a boolean constant is set
// Taken from user comments in PHP documentation for function constant()
function isBooleanConstantAndTrue ($constName) { // : Boolean
	// Failed by default
	$res = false;

	// In cache?
	if (isset($GLOBALS['cache_array']['const'][$constName])) {
		// Use cache
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): " . $constName."-CACHE!<br />");
		$res = ($GLOBALS['cache_array']['const'][$constName] === true);
	} else {
		// Check constant
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): " . $constName."-RESOLVE!<br />");
		if (defined($constName)) {
			// Found!
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): " . $constName."-FOUND!<br />");
			$res = (constant($constName) === true);
		} // END - if

		// Set cache
		$GLOBALS['cache_array']['const'][$constName] = $res;
	}
	//* DEBUG: */ var_dump($res);

	// Return value
	return $res;
}

// Checks if a given apache module is loaded
function isApacheModuleLoaded ($apacheModule) {
	// Check it and return result
	return (((function_exists('apache_get_modules')) && (in_array($apacheModule, apache_get_modules()))) || (!function_exists('apache_get_modules')));
}

// Get current theme name
function getCurrentTheme () {
	// The default theme is 'default'... ;-)
	$ret = 'default';

	// Load default theme if not empty from configuration
	if ((isConfigEntrySet('default_theme')) && (getConfig('default_theme') != '')) $ret = getConfig('default_theme');

	if (!isSessionVariableSet('mxchange_theme')) {
		// Set default theme
		setTheme($ret);
	} elseif ((isSessionVariableSet('mxchange_theme')) && (isExtensionInstalledAndNewer('sql_patches', '0.1.4'))) {
		//die("<pre>".print_r($GLOBALS['cache_array']['themes'], true)."</pre>");
		// Get theme from cookie
		$ret = getSession('mxchange_theme');

		// Is it valid?
		if (getThemeId($ret) == 0) {
			// Fix it to default
			$ret = 'default';
		} // END - if
	} elseif ((!isInstalled()) && ((isInstalling()) || (getOutputMode() == true)) && ((isGetRequestElementSet('theme')) || (isPostRequestElementSet('theme')))) {
		// Prepare FQFN for checking
		$theme = sprintf("%stheme/%s/theme.php", getConfig('PATH'), getRequestElement('theme'));

		// Installation mode active
		if ((isGetRequestElementSet('theme')) && (isFileReadable($theme))) {
			// Set cookie from URL data
			setTheme(getRequestElement('theme'));
		} elseif (isFileReadable(sprintf("%stheme/%s/theme.php", getConfig('PATH'), SQL_ESCAPE(postRequestElement('theme'))))) {
			// Set cookie from posted data
			setTheme(SQL_ESCAPE(postRequestElement('theme')));
		}

		// Set return value
		$ret = getSession('mxchange_theme');
	} else {
		// Invalid design, reset cookie
		setTheme($ret);
	}

	// Return theme value
	return $ret;
}

// Setter for theme in session
function setTheme ($newTheme) {
	setSession('mxchange_theme', $newTheme);
}

// Get id from theme
// @TODO Try to move this to inc/libs/theme_functions.php
function getThemeId ($name) {
	// Is the extension 'theme' installed?
	if (!isExtensionActive('theme')) {
		// Then abort here
		return 0;
	} // END - if

	// Default id
	$id = 0;

	// Is the cache entry there?
	if (isset($GLOBALS['cache_array']['themes']['id'][$name])) {
		// Get the version from cache
		$id = $GLOBALS['cache_array']['themes']['id'][$name];

		// Count up
		incrementStatsEntry('cache_hits');
	} elseif (getExtensionVersion('cache') != '0.1.8') {
		// Check if current theme is already imported or not
		$result = SQL_QUERY_ESC("SELECT `id` FROM `{?_MYSQL_PREFIX?}_themes` WHERE `theme_path`='%s' LIMIT 1",
			array($name), __FUNCTION__, __LINE__);

		// Entry found?
		if (SQL_NUMROWS($result) == 1) {
			// Fetch data
			list($id) = SQL_FETCHROW($result);
		} // END - if

		// Free result
		SQL_FREERESULT($result);
	}

	// Return id
	return $id;
}

// Generates an error code from given account status
function generateErrorCodeFromUserStatus ($status) {
	// @TODO The status should never be empty
	if (empty($status)) {
		// Something really bad happend here
		debug_report_bug(__FUNCTION__ . ': status is empty.');
	} // END - if

	// Default error code if unknown account status
	$errorCode = getCode('UNKNOWN_STATUS');

	// Generate constant name
	$constantName = sprintf("ID_%s", $status);

	// Is the constant there?
	if (isCodeSet($constantName)) {
		// Then get it!
		$errorCode = getCode($constantName);
	} else {
		// Unknown status
		logDebugMessage(__FUNCTION__, __LINE__, sprintf("Unknown error status %s detected.", $status));
	}

	// Return error code
	return $errorCode;
}

// Function to search for the last modifified file
function searchDirsRecursive ($dir, &$last_changed) {
	// Get dir as array
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):dir=" . $dir."<br />");
	// Does it match what we are looking for? (We skip a lot files already!)
	// RegexPattern to exclude  ., .., .revision,  .svn, debug.log or .cache in the filenames
	$excludePattern = '@(\.revision|debug\.log|\.cache|config\.php)$@';
	$ds = getArrayFromDirectory($dir, '', true, false, array(), '.php', $excludePattern);
	//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):ds[]=".count($ds)."<br />");

	// Walk through all entries
	foreach ($ds as $d) {
		// Generate proper FQFN
		$FQFN = str_replace('//', '/', getConfig('PATH') . $dir. '/'. $d);

		// Is it a file and readable?
		//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):FQFN={$FQFN}<br />");
		if (isDirectory($FQFN)) {
			// $FQFN is a directory so also crawl into this directory
			$newDir = $d;
			if (!empty($dir)) $newDir = $dir . '/'. $d;
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):DESCENT: " . $newDir."<br />");
			searchDirsRecursive($newDir, $last_changed);
		} elseif (isFileReadable($FQFN)) {
			// $FQFN is a filename and no directory
			$time = filemtime($FQFN);
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):File: " . $d." found. (".($last_changed['time'] - $time).")<br />");
			if ($last_changed['time'] < $time) {
				// This file is newer as the file before
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>) - NEWER!<br />");
				$last_changed['path_name'] = $FQFN;
				$last_changed['time'] = $time;
			} // END - if
		}
	} // END - foreach
}

// "Getter" for revision/version data
function getActualVersion ($type = 'Revision') {
	// By default nothing is new... ;-)
	$new = false;

	// Is the cache entry there?
	if (isset($GLOBALS['cache_array']['revision'][$type])) {
		// Found so increase cache hit
		incrementStatsEntry('cache_hits');

		// Return it
		return $GLOBALS['cache_array']['revision'][$type][0];
	} else {
		// FQFN of revision file
		$FQFN = sprintf("%s/.revision", getConfig('CACHE_PATH'));

		// Check if 'check_revision_data' is setted (switch for manually rewrite the .revision-File)
		if ((isGetRequestElementSet('check_revision_data')) && (getRequestElement('check_revision_data') == 'yes')) {
			// Forced rebuild of .revision file
			$new = true;
		} else {
			// Check for revision file
			if (!isFileReadable($FQFN)) {
				// Not found, so we need to create it
				$new = true;
			} else {
				// Revision file found
				$ins_vers = explode("\n", readFromFile($FQFN));

				// Get array for mapping information
				$mapper = array_flip(getSearchFor());
				//* DEBUG: */ print('<pre>mapper='.print_r($mapper, true).'</pre>ins_vers=<pre>'.print_r($ins_vers, true).'</pre>');

				// Is the content valid?
				if ((!is_array($ins_vers)) || (count($ins_vers) <= 0) || (!isset($ins_vers[$mapper[$type]])) || (trim($ins_vers[$mapper[$type]]) == '') || ($ins_vers[0]) == 'new') {
					// File needs update!
					$new = true;
				} else {
					// Generate fake cache entry
					foreach ($mapper as $map=>$idx) {
						$GLOBALS['cache_array']['revision'][$map][0] = $ins_vers[$idx];
					} // END - foreach

					// Return found value
					return trim($ins_vers[$mapper[$type]]);
				}
			}
		}

		// Has it been updated?
		if ($new === true)  {
			// Write it
			writeToFile($FQFN, implode("\n", getArrayFromActualVersion()));

			// ... and call recursive
			return getActualVersion($type);
		} // END - if
	}
}

// Repares an array we are looking for
// The returned Array is needed twice (in getArrayFromActualVersion() and in getActualVersion() in the old .revision-fallback) so I puted it in an extra function to not polute the global namespace
function getSearchFor () {
	// Add Revision, Date, Tag and Author
	$searchFor = array('Revision', 'Date', 'Tag', 'Author', 'File');

	// Return the created array
	return $searchFor;
}

// @TODO Please describe this function
function getArrayFromActualVersion () {
	// Init variables
	$next_dir = '';

	// Directory to start with search
	$last_changed = array(
		'path_name' => '',
		'time'      => 0
	);

	// Init return array
	$akt_vers = array();

	// Init value for counting the founded keywords
	$res = 0;

	// Searches all Files and there date of the last modifikation and puts the newest File in $last_changed.
	searchDirsRecursive($next_dir, $last_changed); // @TODO small change to API to $last_changed = searchDirsRecursive($next_dir, $time);

	// Get file
	$last_file = readFromFile($last_changed['path_name']);

	// Get all the keywords to search for
	$searchFor = getSearchFor();

	// This foreach loops the $searchFor-Tags (array('Revision', 'Date', 'Tag', 'Author') --> could easaly extended in the future)
	foreach ($searchFor as $search) {
		// Searches for "$search-tag:VALUE$" or "$search-tag::VALUE$"(the stylish keywordversion ;-)) in the lates modified file
		$res += preg_match('@\$' . $search.'(:|::) (.*) \$@U', $last_file, $t);
		// This trimms the search-result and puts it in the $GLOBALS['cache_array']['revision']-return array
		if (isset($t[2])) $GLOBALS['cache_array']['revision'][$search] = trim($t[2]);
	} // END - foreach

	// Save the last-changed filename for debugging
	$GLOBALS['cache_array']['revision']['File'] = $last_changed['path_name'];

	// at least 3 keyword-Tags are needed for propper values
	if ($res && $res >= 3
	&& isset($GLOBALS['cache_array']['revision']['Revision']) && $GLOBALS['cache_array']['revision']['Revision'] != ''
	&& isset($GLOBALS['cache_array']['revision']['Date']) && $GLOBALS['cache_array']['revision']['Date'] != ''
	&& isset($GLOBALS['cache_array']['revision']['Tag']) && $GLOBALS['cache_array']['revision']['Tag'] != '') {
		// Prepare content witch need special treadment

		// Prepare timestamp for date
		preg_match('@(....)-(..)-(..) (..):(..):(..)@', $GLOBALS['cache_array']['revision']['Date'], $match_d);
		$GLOBALS['cache_array']['revision']['Date'] = mktime($match_d[4], $match_d[5], $match_d[6], $match_d[2], $match_d[3], $match_d[1]);

		// Add author to the Tag if the author is set and is not quix0r (lead coder)
		if ((isset($GLOBALS['cache_array']['revision']['Author'])) && ($GLOBALS['cache_array']['revision']['Author'] != 'quix0r')) {
			$GLOBALS['cache_array']['revision']['Tag'] .= '-'.strtoupper($GLOBALS['cache_array']['revision']['Author']);
		} // END - if

	} else {
		// No valid Data from the last modificated file so read the Revision from the Server. Fallback-solution!! Should not be removed I think.
		$version = sendGetRequest('check-updates3.php');

		// Prepare content
		// Only sets not setted or not proper values to the Online-Server-Fallback-Solution
		if (!isset($GLOBALS['cache_array']['revision']['Revision']) || $GLOBALS['cache_array']['revision']['Revision'] == '') $GLOBALS['cache_array']['revision']['Revision'] = trim($version[10]);
		if (!isset($GLOBALS['cache_array']['revision']['Date'])     || $GLOBALS['cache_array']['revision']['Date']     == '') $GLOBALS['cache_array']['revision']['Date']     = trim($version[9]);
		if (!isset($GLOBALS['cache_array']['revision']['Tag'])      || $GLOBALS['cache_array']['revision']['Tag']      == '') $GLOBALS['cache_array']['revision']['Tag']      = trim($version[8]);
		if (!isset($GLOBALS['cache_array']['revision']['Author'])   || $GLOBALS['cache_array']['revision']['Author']   == '') $GLOBALS['cache_array']['revision']['Author']   = 'quix0r';
		if (!isset($GLOBALS['cache_array']['revision']['File'])     || $GLOBALS['cache_array']['revision']['File']     == '') $GLOBALS['cache_array']['revision']['File']     = trim($version[11]);
	}

	// Return prepared array
	return $GLOBALS['cache_array']['revision'];
}

// Back-ported from the new ship-simu engine. :-)
function debug_get_printable_backtrace () {
	// Init variable
	$backtrace = "<ol>\n";

	// Get and prepare backtrace for output
	$backtraceArray = debug_backtrace();
	foreach ($backtraceArray as $key => $trace) {
		if (!isset($trace['file'])) $trace['file'] = __FUNCTION__;
		if (!isset($trace['line'])) $trace['line'] = __LINE__;
		if (!isset($trace['args'])) $trace['args'] = array();
		$backtrace .= "<li class=\"debug_list\"><span class=\"backtrace_file\">".basename($trace['file'])."</span>:" . $trace['line'].", <span class=\"backtrace_function\">" . $trace['function'].'('.count($trace['args']).")</span></li>\n";
	} // END - foreach

	// Close it
	$backtrace .= "</ol>\n";

	// Return the backtrace
	return $backtrace;
}

// Output a debug backtrace to the user
function debug_report_bug ($message = '') {
	// Is this already called?
	if (isset($GLOBALS[__FUNCTION__])) {
		// Other backtrace
		print 'Message:'.$message.'<br />Backtrace:<pre>';
		debug_print_backtrace();
		die('</pre>');
	} // END - if

	// Set this function as called
	$GLOBALS[__FUNCTION__] = true;

	// Init message
	$debug = '';

	// Is the optional message set?
	if (!empty($message)) {
		// Use and log it
		$debug = sprintf("Note: %s<br />\n",
			$message
		);

		// @TODO Add a little more infos here
		logDebugMessage(__FUNCTION__, __LINE__, strip_tags($message));
	} // END - if

	// Add output
	$debug .= "Please report this bug at <a title=\"Direct link to the bug-tracker\" href=\"http://bugs.mxchange.org\" rel=\"external\" target=\"_blank\">bugs.mxchange.org</a> and include the logfile from <strong>" . getConfig('CACHE_PATH') . "debug.log</strong> in your report (you can now attach files):<pre>";
	$debug .= debug_get_printable_backtrace();
	$debug .= "</pre>\nRequest-URI: " . getRequestUri()."<br />\n";
	$debug .= "Thank you for finding bugs.";

	// And abort here
	// @TODO This cannot be rewritten to app_die(), try to find a solution for this.
	die($debug);
}

// Generates a ***weak*** seed (taken from de.php.net/mt_srand)
function generateSeed () {
	list($usec, $sec) = explode(' ', microtime());
	$microTime = (((float)$sec + (float)$usec)) * 100000;
	return $microTime;
}

// Converts a message code to a human-readable message
function getMessageFromErrorCode ($code) {
	$message = '';
	switch ($code) {
		case '': break;
		case getCode('LOGOUT_DONE')      : $message = getMessage('LOGOUT_DONE'); break;
		case getCode('LOGOUT_FAILED')    : $message = "<span class=\"guest_failed\">{--LOGOUT_FAILED--}</span>"; break;
		case getCode('DATA_INVALID')     : $message = getMessage('MAIL_DATA_INVALID'); break;
		case getCode('POSSIBLE_INVALID') : $message = getMessage('MAIL_POSSIBLE_INVALID'); break;
		case getCode('ACCOUNT_LOCKED')   : $message = getMessage('MEMBER_ACCOUNT_LOCKED_UNC'); break;
		case getCode('USER_404')         : $message = getMessage('USER_404'); break;
		case getCode('STATS_404')        : $message = getMessage('MAIL_STATS_404'); break;
		case getCode('ALREADY_CONFIRMED'): $message = getMessage('MAIL_ALREADY_CONFIRMED'); break;
		case getCode('WRONG_PASS')       : $message = getMessage('LOGIN_WRONG_PASS'); break;
		case getCode('WRONG_ID')         : $message = getMessage('LOGIN_WRONG_ID'); break;
		case getCode('ID_LOCKED')        : $message = getMessage('LOGIN_ID_LOCKED'); break;
		case getCode('ID_UNCONFIRMED')   : $message = getMessage('LOGIN_ID_UNCONFIRMED'); break;
		case getCode('NO_COOKIES')       : $message = getMessage('LOGIN_NO_COOKIES'); break;
		case getCode('COOKIES_DISABLED') : $message = getMessage('LOGIN_NO_COOKIES'); break;
		case getCode('BEG_SAME_AS_OWN')  : $message = getMessage('BEG_SAME_UID_AS_OWN'); break;
		case getCode('LOGIN_FAILED')     : $message = getMessage('LOGIN_FAILED_GENERAL'); break;
		case getCode('MODULE_MEM_ONLY')  : $message = sprintf(getMessage('MODULE_MEM_ONLY'), getRequestElement('mod')); break;
		case getCode('OVERLENGTH')       : $message = getMessage('MEMBER_TEXT_OVERLENGTH'); break;
		case getCode('URL_FOUND')        : $message = getMessage('MEMBER_TEXT_CONTAINS_URL'); break;
		case getCode('SUBJ_URL')         : $message = getMessage('MEMBER_SUBJ_CONTAINS_URL'); break;
		case getCode('BLIST_URL')        : $message = "{--MEMBER_URL_BLACK_LISTED--}<br />\n{--MEMBER_BLIST_TIME--}: ".generateDateTime(getRequestElement('blist'), 0); break;
		case getCode('NO_RECS_LEFT')     : $message = getMessage('MEMBER_SELECTED_MORE_RECS'); break;
		case getCode('INVALID_TAGS')     : $message = getMessage('MEMBER_HTML_INVALID_TAGS'); break;
		case getCode('MORE_POINTS')      : $message = getMessage('MEMBER_MORE_POINTS_NEEDED'); break;
		case getCode('MORE_RECEIVERS1')  : $message = getMessage('MEMBER_ENTER_MORE_RECEIVERS'); break;
		case getCode('MORE_RECEIVERS2')  : $message = getMessage('MEMBER_NO_MORE_RECEIVERS_FOUND'); break;
		case getCode('MORE_RECEIVERS3')  : $message = sprintf(getMessage('MEMBER_ENTER_MORE_MIN_RECEIVERS'), getConfig('order_min')); break;
		case getCode('INVALID_URL')      : $message = getMessage('MEMBER_ENTER_INVALID_URL'); break;

		case getCode('ERROR_MAILID'):
			if (isExtensionActive('mailid', true)) {
				$message = getMessage('ERROR_CONFIRMING_MAIL');
			} else {
				$message = sprintf(getMessage('EXTENSION_PROBLEM_NOT_INSTALLED'), 'mailid');
			}
			break;

		case getCode('EXTENSION_PROBLEM'):
			if (isGetRequestElementSet('ext')) {
				$message = generateExtensionInactiveNotInstalledMessage(getRequestElement('ext'));
			} else {
				$message = getMessage('EXTENSION_PROBLEM_UNSET_EXT');
			}
			break;

		case getCode('URL_TLOCK'):
			$result = SQL_QUERY_ESC("SELECT `timestamp` FROM `{?_MYSQL_PREFIX?}_pool` WHERE `id`=%s LIMIT 1",
				array(bigintval(getRequestElement('id'))), __FILE__, __LINE__);

			// Load timestamp from last order
			list($timestamp) = SQL_FETCHROW($result);
			$timestamp = generateDateTime($timestamp, 1);

			// Free memory
			SQL_FREERESULT($result);

			// Calculate hours...
			$STD = round(getConfig('url_tlock') / 60 / 60);

			// Minutes...
			$MIN = round((getConfig('url_tlock') - $STD * 60 * 60) / 60);

			// And seconds
			$SEC = getConfig('url_tlock') - $STD * 60 * 60 - $MIN * 60;

			// Finally contruct the message
			// @TODO Rewrite this old lost code to a template
			$message = "{--MEMBER_URL_TIME_LOCK--}<br />{--CONFIG_URL_TLOCK--} ".$STD."
			{--_HOURS--}, ".$MIN." {--_MINUTES--} {--_AND--} ".$SEC." {--_SECONDS--}<br />
			{--MEMBER_LAST_TLOCK--}: ".$timestamp;
			break;

		default:
			// Missing/invalid code
			$message = sprintf(getMessage('UNKNOWN_MAILID_CODE'), $code);

			// Log it
			logDebugMessage(__FUNCTION__, __LINE__, $message);
			break;
	} // END - switch

	// Return the message
	return $message;
}

// Generate a "link" for the given admin id (admin_id)
function generateAdminLink ($adminId) {
	// No assigned admin is default
	$admin = "<span class=\"admin_note\">{--ADMIN_NO_ADMIN_ASSIGNED--}</span>";

	// Zero? = Not assigned
	if (bigintval($adminId) > 0) {
		// Load admin's login
		$login = getAdminLogin($adminId);

		// Is the login valid?
		if ($login != '***') {
			// Is the extension there?
			if (isExtensionActive('admins')) {
				// Admin found
				$admin = "<a href=\"".generateEmailLink(getAdminEmail($adminId), 'admins')."\">" . $login."</a>";
			} else {
				// Extension not found
				$admin = sprintf(getMessage('EXTENSION_PROBLEM_NOT_INSTALLED'), 'admins');
			}
		} else {
			// Maybe deleted?
			$admin = "<div class=\"admin_note\">".sprintf(getMessage('ADMIN_ID_404'), $adminId)."</div>";
		}
	} // END - if

	// Return result
	return $admin;
}

// Compile characters which are allowed in URLs
function compileUriCode ($code, $simple = true) {
	// Compile constants
	if ($simple === false) $code = str_replace('{--', '".', str_replace('--}', '."', $code));

	// Compile QUOT and other non-HTML codes
	$code = str_replace('{DOT}', '.',
		str_replace('{SLASH}', '/',
		str_replace('{QUOT}', "'",
		str_replace('{DOLLAR}', '$',
		str_replace('{OPEN_ANCHOR}', '(',
		str_replace('{CLOSE_ANCHOR}', ')',
		str_replace('{OPEN_SQR}', '[',
		str_replace('{CLOSE_SQR}', ']',
		str_replace('{PER}', '%',
		$code
	)))))))));

	// Return compiled code
	return $code;
}

// Function taken from user comments on www.php.net / function eregi()
function isUrlValidSimple ($url) {
	// Prepare URL
	$url = secureString(str_replace("\\", '', compileCode(urldecode($url))));

	// Allows http and https
	$http      = "(http|https)+(:\/\/)";
	// Test domain
	$domain1   = "([[:alnum:]]([-[:alnum:]])*\.)?([[:alnum:]][-[:alnum:]\.]*[[:alnum:]])(\.[[:alpha:]]{2,5})?";
	// Test double-domains (e.g. .de.vu)
	$domain2   = "([-[:alnum:]])?(\.[[:alnum:]][-[:alnum:]\.]*[[:alnum:]])(\.[[:alpha:]]{2,5})(\.[[:alpha:]]{2,5})?";
	// Test IP number
	$ip        = "([[:digit:]]{1,3})\.([[:digit:]]{1,3})\.([[:digit:]]{1,3})\.([[:digit:]]{1,3})";
	// ... directory
	$dir       = "((/)+([-_\.[:alnum:]])+)*";
	// ... page
	$page      = "/([-_[:alnum:]][-\._[:alnum:]]*\.[[:alnum:]]{2,5})?";
	// ... and the string after and including question character
	$getstring1 = "([\?/]([[:alnum:]][-\._%[:alnum:]]*(=)?([-\@\._:%[:alnum:]])+)(&([[:alnum:]]([-_%[:alnum:]])*(=)?([-\@\[\._:%[:alnum:]])+(\])*))*)?";
	// Pattern for URLs like http://url/dir/doc.html?var=value
	$pattern['d1dpg1']  = $http . $domain1 . $dir . $page . $getstring1;
	$pattern['d2dpg1']  = $http . $domain2 . $dir . $page . $getstring1;
	$pattern['ipdpg1']  = $http . $ip . $dir . $page . $getstring1;
	// Pattern for URLs like http://url/dir/?var=value
	$pattern['d1dg1']  = $http . $domain1 . $dir.'/' . $getstring1;
	$pattern['d2dg1']  = $http . $domain2 . $dir.'/' . $getstring1;
	$pattern['ipdg1']  = $http . $ip . $dir.'/' . $getstring1;
	// Pattern for URLs like http://url/dir/page.ext
	$pattern['d1dp']  = $http . $domain1 . $dir . $page;
	$pattern['d1dp']  = $http . $domain2 . $dir . $page;
	$pattern['ipdp']  = $http . $ip . $dir . $page;
	// Pattern for URLs like http://url/dir
	$pattern['d1d']  = $http . $domain1 . $dir;
	$pattern['d2d']  = $http . $domain2 . $dir;
	$pattern['ipd']  = $http . $ip . $dir;
	// Pattern for URLs like http://url/?var=value
	$pattern['d1g1']  = $http . $domain1 . '/' . $getstring1;
	$pattern['d2g1']  = $http . $domain2 . '/' . $getstring1;
	$pattern['ipg1']  = $http . $ip . '/' . $getstring1;
	// Pattern for URLs like http://url?var=value
	$pattern['d1g12']  = $http . $domain1 . $getstring1;
	$pattern['d2g12']  = $http . $domain2 . $getstring1;
	$pattern['ipg12']  = $http . $ip . $getstring1;
	// Test all patterns
	$reg = false;
	foreach ($pattern as $key => $pat) {
		// Debug regex?
		if (isDebugRegExpressionEnabled()) {
			// @TODO Are these convertions still required?
			$pat = str_replace('.', "&#92;&#46;", $pat);
			$pat = str_replace('@', "&#92;&#64;", $pat);
			//* DEBUG: */ outputHtml($key."=&nbsp;" . $pat . "<br />");
		} // END - if

		// Check if expression matches
		$reg = ($reg || preg_match(('^' . $pat.'^'), $url));

		// Does it match?
		if ($reg === true) break;
	}

	// Return true/false
	return $reg;
}

// Wtites data to a config.php-style file
// @TODO Rewrite this function to use readFromFile() and writeToFile()
function changeDataInFile ($FQFN, $comment, $prefix, $suffix, $DATA, $seek=0) {
	// Initialize some variables
	$done = false;
	$seek++;
	$next  = -1;
	$found = false;

	// Is the file there and read-/write-able?
	if ((isFileReadable($FQFN)) && (is_writeable($FQFN))) {
		$search = 'CFG: ' . $comment;
		$tmp = $FQFN . '.tmp';

		// Open the source file
		$fp = fopen($FQFN, 'r') or outputHtml('<strong>READ:</strong> ' . $FQFN . '<br />');

		// Is the resource valid?
		if (is_resource($fp)) {
			// Open temporary file
			$fp_tmp = fopen($tmp, 'w') or outputHtml('<strong>WRITE:</strong> ' . $tmp . '<br />');

			// Is the resource again valid?
			if (is_resource($fp_tmp)) {
				// Mark temporary file as readable
				$GLOBALS['file_readable'][$tmp] = true;

				// Start reading
				while (!feof($fp)) {
					// Read from source file
					$line = fgets ($fp, 1024);

					if (strpos($line, $search) > -1) { $next = 0; $found = true; }

					if ($next > -1) {
						if ($next === $seek) {
							$next = -1;
							$line = $prefix . $DATA . $suffix . "\n";
						} else {
							$next++;
						}
					} // END - if

					// Write to temp file
					fputs($fp_tmp, $line);
				} // END - while

				// Close temp file
				fclose($fp_tmp);

				// Finished writing tmp file
				$done = true;
			} // END - if

			// Close source file
			fclose($fp);

			if (($done === true) && ($found === true)) {
				// Copy back tmp file and delete tmp :-)
				copyFileVerified($tmp, $FQFN, 0644);
				return removeFile($tmp);
			} elseif ($found === false) {
				outputHtml('<strong>CHANGE:</strong> 404!');
			} else {
				outputHtml('<strong>TMP:</strong> UNDONE!');
			}
		}
	} else {
		// File not found, not readable or writeable
		outputHtml('<strong>404:</strong> ' . $FQFN . '<br />');
	}

	// An error was detected!
	return false;
}
// Send notification to admin
function sendAdminNotification ($subject, $templateName, $content=array(), $userid = 0) {
	if (getExtensionVersion('admins') >= '0.4.1') {
		// Send new way
		sendAdminsEmails($subject, $templateName, $content, $userid);
	} else {
		// Send out out-dated way
		$message = loadEmailTemplate($templateName, $content, $userid);
		sendAdminEmails($subject, $message);
	}
}

// Debug message logger
function logDebugMessage ($funcFile, $line, $message, $force=true) {
	// Is debug mode enabled?
	if ((isDebugModeEnabled()) || ($force === true)) {
		// Remove CRLF
		$message = str_replace("\r", '', str_replace("\n", '', $message));

		// Log this message away, we better don't call app_die() here to prevent an endless loop
		$fp = fopen(getConfig('CACHE_PATH') . 'debug.log', 'a') or die(__FUNCTION__.'['.__LINE__.']: Cannot write logfile debug.log!');
		fwrite($fp, date('d.m.Y|H:i:s', time()) . '|' . getModule() . '|' . basename($funcFile) . '|' . $line . '|' . $message . "\n");
		fclose($fp);
	} // END - if
}

// Handle extra values
function handleExtraValues ($filterFunction, $value, $extraValue) {
	// Default is the value itself
	$ret = $value;

	// Do we have a special filter function?
	if (!empty($filterFunction)) {
		// Does the filter function exist?
		if (function_exists($filterFunction)) {
			// Do we have extra parameters here?
			if (!empty($extraValue)) {
				// Put both parameters in one new array by default
				$args = array($value, $extraValue);

				// If we have an array simply use it and pre-extend it with our value
				if (is_array($extraValue)) {
					// Make the new args array
					$args = merge_array(array($value), $extraValue);
				} // END - if

				// Call the multi-parameter call-back
				$ret = call_user_func_array($filterFunction, $args);
			} else {
				// One parameter call
				$ret = call_user_func($filterFunction, $value);
			}
		} // END - if
	} // END - if

	// Return the value
	return $ret;
}

// Converts timestamp selections into a timestamp
function convertSelectionsToTimestamp (&$postData, &$DATA, &$id, &$skip) {
	// Init test variable
	$test2 = '';

	// Get last three chars
	$test = substr($id, -3);

	// Improved way of checking! :-)
	if (in_array($test, array('_ye', '_mo', '_we', '_da', '_ho', '_mi', '_se'))) {
		// Found a multi-selection for timings?
		$test = substr($id, 0, -3);
		if ((isset($postData[$test.'_ye'])) && (isset($postData[$test.'_mo'])) && (isset($postData[$test.'_we'])) && (isset($postData[$test.'_da'])) && (isset($postData[$test.'_ho'])) && (isset($postData[$test.'_mi'])) && (isset($postData[$test.'_se'])) && ($test != $test2)) {
			// Generate timestamp
			$postData[$test] = createTimestampFromSelections($test, $postData);
			$DATA[] = sprintf("%s='%s'", $test, $postData[$test]);

			// Remove data from array
			foreach (array('ye', 'mo', 'we', 'da', 'ho', 'mi', 'se') as $rem) {
				unset($postData[$test.'_' . $rem]);
			} // END - foreach

			// Skip adding
			unset($id); $skip = true; $test2 = $test;
		} // END - if
	} else {
		// Process this entry
		$skip = false;
		$test2 = '';
	}
}

// Reverts the german decimal comma into Computer decimal dot
function convertCommaToDot ($str) {
	// Default float is not a float... ;-)
	$float = false;

	// Which language is selected?
	switch (getLanguage()) {
		case 'de': // German language
			// Remove german thousand dots first
			$str = str_replace('.', '', $str);

			// Replace german commata with decimal dot and cast it
			$float = (float)str_replace(',', '.', $str);
			break;

		default: // US and so on
			// Remove thousand dots first and cast
			$float = (float)str_replace(',', '', $str);
			break;
	}

	// Return float
	return $float;
}

// Handle menu-depending failed logins and return the rendered content
function handleLoginFailtures ($accessLevel) {
	// Default output is empty ;-)
	$OUT = '';

	// Is the session data set?
	if ((isSessionVariableSet('mxchange_' . $accessLevel.'_failures')) && (isSessionVariableSet('mxchange_' . $accessLevel.'_last_fail'))) {
		// Ignore zero values
		if (getSession('mxchange_' . $accessLevel.'_failures') > 0) {
			// Non-guest has login failures found, get both data and prepare it for template
			//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>):accessLevel={$accessLevel}<br />");
			$content = array(
				'login_failures' => getSession('mxchange_' . $accessLevel.'_failures'),
				'last_failure'   => generateDateTime(getSession('mxchange_' . $accessLevel.'_last_fail'), 2)
			);

			// Load template
			$OUT = loadTemplate('login_failures', true, $content);
		} // END - if

		// Reset session data
		setSession('mxchange_' . $accessLevel.'_failures', '');
		setSession('mxchange_' . $accessLevel.'_last_fail', '');
	} // END - if

	// Return rendered content
	return $OUT;
}

// Rebuild cache
function rebuildCacheFile ($cache, $inc = '', $force = false) {
	// Debug message
	/* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("cache=%s, inc=%s, force=%s", $cache, $inc, intval($force)));

	// Shall I remove the cache file?
	if (isCacheInstanceValid()) {
		// Rebuild cache
		if ($GLOBALS['cache_instance']->loadCacheFile($cache)) {
			// Destroy it
			$GLOBALS['cache_instance']->removeCacheFile($force);
		} // END - if

		// Include file given?
		if (!empty($inc)) {
			// Construct FQFN
			$inc = sprintf("inc/loader/load_cache-%s.php", $inc);

			// Is the include there?
			if (isIncludeReadable($inc)) {
				// And rebuild it from scratch
				//* DEBUG: */ outputHtml(__FUNCTION__."(<font color=\"#0000aa\">".__LINE__."</font>): inc={$inc} - LOADED!<br />");
				loadInclude($inc);
			} else {
				// Include not found!
				logDebugMessage(__FUNCTION__, __LINE__, "Include {$inc} not found. cache={$cache}");
			}
		} // END - if
	} // END - if
}

// Determines the real remote address
function determineRealRemoteAddress () {
	// Is a proxy in use?
	if (isset($_SERVER['HTTP_X_FORWARDED_FOR'])) {
		// Proxy was used
		$address = $_SERVER['HTTP_X_FORWARDED_FOR'];
	} elseif (isset($_SERVER['HTTP_CLIENT_IP'])) {
		// Yet, another proxy
		$address = $_SERVER['HTTP_CLIENT_IP'];
	} else {
		// The regular address when no proxy was used
		$address = $_SERVER['REMOTE_ADDR'];
	}

	// This strips out the real address from proxy output
	if (strstr($address, ',')) {
		$addressArray = explode(',', $address);
		$address = $addressArray[0];
	} // END - if

	// Return the result
	return $address;
}

// Adds a bonus mail to the queue
// This is a high-level function!
function addNewBonusMail ($data, $mode = '', $output=true) {
	// Use mode from data if not set and availble ;-)
	if ((empty($mode)) && (isset($data['mode']))) $mode = $data['mode'];

	// Generate receiver list
	$RECEIVER = generateReceiverList($data['cat'], $data['receiver'], $mode);

	// Receivers added?
	if (!empty($RECEIVER)) {
		// Add bonus mail to queue
		addBonusMailToQueue(
		$data['subject'],
		$data['text'],
		$RECEIVER,
		$data['points'],
		$data['seconds'],
		$data['url'],
		$data['cat'],
		$mode,
		$data['receiver']
		);

		// Mail inserted into bonus pool
		if ($output) loadTemplate('admin_settings_saved', false, getMessage('ADMIN_BONUS_SEND'));
	} elseif ($output) {
		// More entered than can be reached!
		loadTemplate('admin_settings_saved', false, getMessage('ADMIN_MORE_SELECTED'));
	} else {
		// Debug log
		logDebugMessage(__FUNCTION__, __LINE__, "cat={$data['cat']},receiver={$data['receiver']},data=".base64_encode(serialize($data))." More selected, than available!");
	}
}

// Determines referal id and sets it
function determineReferalId () {
	// Skip this in non-html-mode
	if (getOutputMode() != 0) return false;

	// Check if refid is set
	if ((isset($GLOBALS['refid'])) && ($GLOBALS['refid'] > 0)) {
		// This is fine...
	} elseif ((isGetRequestElementSet('user')) && (basename($_SERVER['PHP_SELF']) == 'click.php')) {
		// The variable user comes from the click-counter script click.php and we only accept this here
		$GLOBALS['refid'] = bigintval(getRequestElement('user'));
	} elseif (isPostRequestElementSet('refid')) {
		// Get referal id from variable refid (so I hope this makes my script more compatible to other scripts)
		$GLOBALS['refid'] = secureString(postRequestElement('refid'));
	} elseif (isGetRequestElementSet('refid')) {
		// Get referal id from variable refid (so I hope this makes my script more compatible to other scripts)
		$GLOBALS['refid'] = secureString(getRequestElement('refid'));
	} elseif (isGetRequestElementSet('ref')) {
		// Set refid=ref (the referal link uses such variable)
		$GLOBALS['refid'] = secureString(getRequestElement('ref'));
	} elseif ((isSessionVariableSet('refid')) && (getSession('refid') != 0)) {
		// Set session refid als global
		$GLOBALS['refid'] = bigintval(getSession('refid'));
	} elseif ((isExtensionInstalledAndNewer('user', '0.3.4')) && (getConfig('select_user_zero_refid')) == 'Y') {
		// Select a random user which has confirmed enougth mails
		$GLOBALS['refid'] = determineRandomReferalId();
	} elseif ((isExtensionInstalled('sql_patches')) && (getConfig('def_refid') > 0)) {
		// Set default refid as refid in URL
		$GLOBALS['refid'] = getConfig('def_refid');
	} else {
		// No default ID when sql_patches is not installed or none set
		$GLOBALS['refid'] = 0;
	}

	// Set cookie when default refid > 0
	if (!isSessionVariableSet('refid') || (!empty($GLOBALS['refid'])) || ((getSession('refid') == 0) && (isConfigEntrySet('def_refid')) && (getConfig('def_refid') > 0))) {
		// Set cookie
		setSession('refid', $GLOBALS['refid']);
	} // END - if

	// Return determined refid
	return $GLOBALS['refid'];
}

// Enables the reset mode and runs it
function doReset () {
	// Enable the reset mode
	$GLOBALS['reset_enabled'] = true;

	// Run filters
	runFilterChain('reset');
}

// Our shutdown-function
function shutdown () {
	// Call the filter chain 'shutdown'
	runFilterChain('shutdown', null);

	if (SQL_IS_LINK_UP()) {
		// Close link
		SQL_CLOSE(__FILE__, __LINE__);
	} elseif (!isInstallationPhase()) {
		// No database link
		addFatalMessage(__FILE__, __LINE__, getMessage('NO_DB_LINK_SHUTDOWN'));
	}

	// Stop executing here
	exit;
}

// Setter for userid
function setUserId ($userid) {
	$GLOBALS['userid'] = bigintval($userid);
}

// Getter for userid or returns zero
function getUserId () {
	// Default userid
	$userid = 0;

	// Is the userid set?
	if (isUserIdSet()) {
		// Then use it
		$userid = $GLOBALS['userid'];
	} // END - if

	// Return it
	return $userid;
}

// Checks ether the userid is set
function isUserIdSet () {
	return (isset($GLOBALS['userid']));
}

// Handle message codes from URL
function handleCodeMessage () {
	if (isGetRequestElementSet('code')) {
		// Default extension is 'unknown'
		$ext = 'unknown';

		// Is extension given?
		if (isGetRequestElementSet('ext')) $ext = getRequestElement('ext');

		// Convert the 'code' parameter from URL to a human-readable message
		$message = getMessageFromErrorCode(getRequestElement('code'));

		// Load message template
		loadTemplate('message', false, $message);
	} // END - if
}

// Setter for extra title
function setExtraTitle ($extraTitle) {
	$GLOBALS['extra_title'] = $extraTitle;
}

// Getter for extra title
function getExtraTitle () {
	// Is the extra title set?
	if (!isExtraTitleSet()) {
		// No, then abort here
		debug_report_bug('extra_title is not set!');
	} // END - if

	// Return it
	return $GLOBALS['extra_title'];
}

// Checks if the extra title is set
function isExtraTitleSet () {
	return ((isset($GLOBALS['extra_title'])) && (!empty($GLOBALS['extra_title'])));
}

// Generates a 'extension foo inactive' message
function generateExtensionInactiveMessage ($ext_name) {
	// Is the extension empty?
	if (empty($ext_name)) {
		// This should not happen
		debug_report_bug(__FUNCTION__ . ': Parameter ext is empty. This should not happen.');
	} // END - if

	// Default message
	$message = sprintf(getMessage('EXTENSION_PROBLEM_EXT_INACTIVE'), $ext_name);

	// Is an admin logged in?
	if (isAdmin()) {
		// Then output admin message
		$message = sprintf(getMessage('ADMIN_EXTENSION_PROBLEM_EXT_INACTIVE'), $ext_name);
	} // END - if

	// Return prepared message
	return $message;
}

// Generates a 'extension foo not installed' message
function generateExtensionNotInstalledMessage ($ext_name) {
	// Is the extension empty?
	if (empty($ext_name)) {
		// This should not happen
		debug_report_bug(__FUNCTION__ . ': Parameter ext is empty. This should not happen.');
	} // END - if

	// Default message
	$message = sprintf(getMessage('EXTENSION_PROBLEM_EXT_NOT_INSTALLED'), $ext_name);

	// Is an admin logged in?
	if (isAdmin()) {
		// Then output admin message
		$message = sprintf(getMessage('ADMIN_EXTENSION_PROBLEM_EXT_NOT_INSTALLED'), $ext_name);
	} // END - if

	// Return prepared message
	return $message;
}

// Generates a message depending on if the extension is not installed or not
// just activated
function generateExtensionInactiveNotInstalledMessage ($ext_name) {
	// Init message
	$message = '';

	// Is the extension not installed or just deactivated?
	switch (isExtensionInstalled($ext_name)) {
		case true; // Deactivated!
			$message = generateExtensionInactiveMessage($ext_name);
			break;

		case false; // Not installed!
			$message = generateExtensionNotInstalledMessage($ext_name);
			break;

		default: // Should not happen!
			logDebugMessage(__FUNCTION__, __LINE__, sprintf("Invalid state of extension %s detected.", $ext_name));
			$message = sprintf("Invalid state of extension %s detected.", $ext_name);
			break;
	} // END - switch

	// Return the message
	return $message;
}

// Reads a directory recursively by default and searches for files not matching
// an exclusion pattern. You can now keep the exclusion pattern empty for reading
// a whole directory.
function getArrayFromDirectory ($baseDir, $prefix, $fileIncludeDirs = false, $addBaseDir = true, $excludeArray = array(), $extension = '.php', $excludePattern = '@(\.|\.\.)$@', $recursive = true) {
	// Add default entries we should exclude
	$excludeArray[] = '.';
	$excludeArray[] = '..';
	$excludeArray[] = '.svn';
	$excludeArray[] = '.htaccess';

	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, "baseDir={$baseDir},prefix={$prefix} - Entered!");
	// Init includes
	$files = array();

	// Open directory
	$dirPointer = opendir(getConfig('PATH') . $baseDir) or app_die(__FUNCTION__, __LINE__, 'Cannot read directory ' . basename($baseDir) . '.');

	// Read all entries
	while ($baseFile = readdir($dirPointer)) {
		// Exclude '.', '..' and entries in $excludeArray automatically
		if (in_array($baseFile, $excludeArray, true))  {
			// Exclude them
			//* DEBUG: */ outputHtml('excluded=' . $baseFile . "<br />");
			continue;
		} // END - if

		// Construct include filename and FQFN
		$fileName = $baseDir . $baseFile;
		$FQFN = getConfig('PATH') . $fileName;

		// Remove double slashes
		$FQFN = str_replace('//', '/', $FQFN);

		// Check if the base filename matches an exclusion pattern and if the pattern is not empty
		if ((!empty($excludePattern)) && (preg_match($excludePattern, $baseFile, $match))) {
			// These Lines are only for debugging!!
			//* DEBUG: */ outputHtml('baseDir:' . $baseDir . "<br />");
			//* DEBUG: */ outputHtml('baseFile:' . $baseFile . "<br />");
			//* DEBUG: */ outputHtml('FQFN:' . $FQFN . "<br />");

			// Exclude this one
			continue;
		} // END - if

		// Skip also files with non-matching prefix genericly
		if (($recursive === true) && (isDirectory($FQFN))) {
			// Is a redirectory so read it as well
			$files = merge_array($files, getArrayFromDirectory($baseDir . $baseFile . '/', $prefix, $fileIncludeDirs, $addBaseDir, $excludeArray, $extension, $excludePattern, $recursive));

			// And skip further processing
			continue;
		} elseif (substr($baseFile, 0, strlen($prefix)) != $prefix) {
			// Skip this file
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, "Invalid prefix in file " . $baseFile . ", prefix=" . $prefix);
			continue;
		} elseif (!isFileReadable($FQFN)) {
			// Not readable so skip it
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, "File " . $FQFN . " is not readable!");
			continue;
		}

		// Is the file a PHP script or other?
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, "baseDir={$baseDir},prefix={$prefix},baseFile={$baseFile}");
		if ((substr($baseFile, -4, 4) == '.php') || (($fileIncludeDirs === true) && (isDirectory($FQFN)))) {
			// Is this a valid include file?
			if ($extension == '.php') {
				// Remove both for extension name
				$extName = substr($baseFile, strlen($prefix), -4);

				// Is the extension valid and active?
				if (isExtensionNameValid($extName)) {
					// Then add this file
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Extension entry ' . $baseFile . ' added.');
					$files[] = $fileName;
				} else {
					// Add non-extension files as well
					//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Regular entry ' . $baseFile . ' added.');
					if ($addBaseDir === true) {
						$files[] = $fileName;
					} else {
						$files[] = $baseFile;
					}
				}
			} else {
				// We found .php file but should not search for them, why?
				debug_report_bug('We should find files with extension=' . $extension . ', but we found a PHP script.');
			}
		} elseif (substr($baseFile, -4, 4) == $extension) {
			// Other, generic file found
			$files[] = $fileName;
		}
	} // END - while

	// Close directory
	closedir($dirPointer);

	// Sort array
	asort($files);

	// Return array with include files
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, '- Left!');
	return $files;
}

// Maps a module name into a database table name
function mapModuleToTable ($moduleName) {
	// Map only these, still lame code...
	switch ($moduleName) {
		// 'index' is the guest's menu
		case 'index': $moduleName = 'guest';  break;
		// ... and 'login' the member's menu
		case 'login': $moduleName = 'member'; break;
		// Anything else will not be mapped, silently.
	} // END - switch

	// Return result
	return $moduleName;
}

// Add SQL debug data to array for later output
function addSqlToDebug ($result, $sqlString, $timing, $F, $L) {
	// Already executed?
	if (isset($GLOBALS['debug_sqls'][$F][$L][$sqlString])) {
		// Then abort here, we don't need to profile a query twice
		return;
	} // END - if

	// Remeber this as profiled (or not, but we don't care here)
	$GLOBALS['debug_sqls'][$F][$L][$sqlString] = true;

	// Do we have cache?
	if (!isset($GLOBALS['debug_sql_available'])) {
		// Check it and cache it in $GLOBALS
		$GLOBALS['debug_sql_available'] = ((isExtensionInstalledAndNewer('other', '0.2.2')) && (getConfig('display_debug_sqls') == 'Y'));
	} // END - if
	
	// Don't execute anything here if we don't need or ext-other is missing
	if ($GLOBALS['debug_sql_available'] === false) {
		return;
	} // END - if

	// Generate record
	$record = array(
		'num_rows' => SQL_NUMROWS($result),
		'affected' => SQL_AFFECTEDROWS(),
		'sql_str'  => $sqlString,
		'timing'   => $timing,
		'file'     => basename($F),
		'line'     => $L
	);

	// Add it
	$GLOBALS['debug_sqls'][] = $record;
}

// Initializes the cache instance
function initCacheInstance () {
	// Load include for CacheSystem class
	loadIncludeOnce('inc/classes/cachesystem.class.php');

	// Initialize cache system only when it's needed
	$GLOBALS['cache_instance'] = new CacheSystem();
	if ($GLOBALS['cache_instance']->getStatus() != 'done') {
		// Failed to initialize cache sustem
		addFatalMessage(__FILE__, __LINE__, "(<font color=\"#0000aa\">".__LINE__."</font>): ".getMessage('CACHE_CANNOT_INITIALIZE'));
	} // END - if
}

// Getter for message from array or raw message
function getMessageFromIndexedArray ($message, $pos, $array) {
	// Check if the requested message was found in array
	if (isset($array[$pos])) {
		// ... if yes then use it!
		$ret = $array[$pos];
	} else {
		// ... else use default message
		$ret = $message;
	}

	// Return result
	return $ret;
}

// Print code with line numbers
function linenumberCode ($code)    {
	if (!is_array($code)) $codeE = explode("\n", $code); else $codeE = $code;
	$count_lines = count($codeE);

	$r = 'Line | Code:<br />';
	foreach($codeE as $line => $c) {
		$r .= '<div class="line"><span class="linenum">';
		if ($count_lines == 1) {
			$r .= 1;
		} else {
			$r .= ($line == ($count_lines - 1)) ? '' :  ($line+1);
		}
		$r .= '</span>|';

		// Add code
		$r .= '<span class="linetext">' . htmlentities($c) . '</span></div>';
	}

	return '<div class="code">' . $r . '</div>';
}

// Convert ';' to ', ' for e.g. receiver list
function convertReceivers ($old) {
	return str_replace(';', ', ', $old);
}

//////////////////////////////////////////////////
// AUTOMATICALLY RE-GENERATED MISSING FUNCTIONS //
//////////////////////////////////////////////////
//
if (!function_exists('html_entity_decode')) {
	// Taken from documentation on www.php.net
	function html_entity_decode ($string) {
		$trans_tbl = get_html_translation_table(HTML_ENTITIES);
		$trans_tbl = array_flip($trans_tbl);
		return strtr($string, $trans_tbl);
	}
} // END - if

if (!function_exists('http_build_query')) {
	// Taken from documentation on www.php.net, credits to Marco K. (Germany)
	function http_build_query($data, $prefix='', $sep='', $key='') {
		$ret = array();
		foreach ((array)$data as $k => $v) {
			if (is_int($k) && $prefix != null) {
				$k = urlencode($prefix . $k);
			} // END - if

			if ((!empty($key)) || ($key === 0))  $k = $key.'['.urlencode($k).']';

			if (is_array($v) || is_object($v)) {
				array_push($ret, http_build_query($v, '', $sep, $k));
			} else {
				array_push($ret, $k.'='.urlencode($v));
			}
		} // END - foreach

		if (empty($sep)) $sep = ini_get('arg_separator.output');

		return implode($sep, $ret);
	}
}// // END - if

// [EOF]
?>
